"""
This file contains functions to read the config file of wrfrun
"""

from copy import deepcopy
from os import environ, makedirs
from os.path import abspath, basename, dirname, exists
from shutil import copyfile
from typing import Optional, Tuple, Union

import f90nml
import tomli
import tomli_w

from .error import ResourceURIError, WRFRunContextError, ModelNameError
from ..utils import logger


class _WRFRunResources:
    """
    Manage resource files used by wrfrun components.
    These resources include various configuration files from NWP as well as those provided by wrfrun itself.
    Since their actual file paths may vary depending on the wrfrun installation environment, wrfrun maps them using URIs to ensure consistent access regardless of the environment.
    """
    def __init__(self):
        self._resource_namespace_db = {}

    def check_resource_uri(self, unique_uri: str) -> bool:
        """
        Check if the uri has been registered.

        :param unique_uri: Unique URI represents the resource.
        :type unique_uri: str
        :return: True or False.
        :rtype: bool
        """
        if unique_uri in self._resource_namespace_db:
            return True

        else:
            return False

    def register_resource_uri(self, unique_uri: str, res_space_path: str):
        """
        This function should only be used by wrfrun functions.

        Register a unique resource file namespace.

        :param unique_uri: Unique URI represents the resource. It must start with ``:WRFRUN_`` and end with ``:``. For example, ``":WRFRUN_WORK_PATH:"``.
        :type unique_uri: str
        :param res_space_path: REAL absolute path of your resource path. For example, "$HOME/.config/wrfrun/res".
        :type res_space_path: str
        :return:
        :rtype:
        """
        if not (unique_uri.startswith(":WRFRUN_") and unique_uri.endswith(":")):
            logger.error(f"Can't register resource URI: '{unique_uri}'. It should start with ':WRFRUN_' and end with ':'.")
            raise ResourceURIError(f"Can't register resource URI: '{unique_uri}'. It should start with ':WRFRUN_' and end with ':'.")

        if unique_uri in self._resource_namespace_db:
            logger.error(f"Resource URI '{unique_uri}' exists.")
            raise ResourceURIError(f"Resource URI '{unique_uri}' exists.")



        logger.debug(f"Register URI '{unique_uri}' to '{res_space_path}'")
        self._resource_namespace_db[unique_uri] = res_space_path

    def parse_resource_uri(self, file_path: str) -> str:
        """
        Return a real file path by parsing the URI string in it.

        Normal path will be returned with no change.

        :param file_path: File path string which may contain URI string.
        :type file_path: str
        :return: Real file path.
        :rtype: str
        """
        if not file_path.startswith(":WRFRUN_"):
            return file_path

        res_namespace_string = file_path.split(":")[1]
        res_namespace_string = f":{res_namespace_string}:"

        if res_namespace_string in self._resource_namespace_db:
            file_path = file_path.replace(res_namespace_string, self._resource_namespace_db[res_namespace_string])

            if not file_path.startswith(":WRFRUN_"):
                return file_path

            else:
                return self.parse_resource_uri(file_path)

        else:
            logger.error(f"Unknown resource URI: '{res_namespace_string}'")
            raise ResourceURIError(f"Unknown resource URI: '{res_namespace_string}'")


class _WRFRunConstants:
    """
    Define all variables that will be used by other wrfrun components.
    These variables are related to the wrfrun installation environment and configuration files.
    They are defined either directly or mapped using URIs to ensure consistent access across all components.
    """
    def __init__(self):
        # the path we may need to store temp files,
        # don't worry, it will be deleted once the system reboots
        self._WRFRUN_TEMP_PATH = "/tmp/wrfrun"

        # WRF may need a large disk space to store output, we can't run wrf in /tmp,
        # so we will create a folder in $HOME/.config to run wrf.
        # we need to check if we're running as a root user
        USER_HOME_PATH = f"{environ['HOME']}"
        if USER_HOME_PATH in ["/", "/root", ""]:
            logger.warning(f"User's home path is '{USER_HOME_PATH}', which means you are running this program as a root user")
            logger.warning("It's not recommended to use wrfrun as a root user")
            logger.warning("Set USER_HOME_PATH as /root")
            USER_HOME_PATH = "/root"

        self._WRFRUN_HOME_PATH = f"{USER_HOME_PATH}/.config/wrfrun"
        self._WRFRUN_REPLAY_WORK_PATH = f"{self._WRFRUN_HOME_PATH}/replay"

        # work path to run WPS, WRF and WRFDA
        self._WORK_PATH = f"{self._WRFRUN_HOME_PATH}/workspace"
        self._WPS_WORK_PATH = f"{self._WORK_PATH}/WPS"
        self._WRF_WORK_PATH = f"{self._WORK_PATH}/WRF"
        self._WRFDA_WORK_PATH = f"{self._WORK_PATH}/WRFDA"

        # record WRF progress status
        self._WORK_STATUS = ""

        # record context status
        self._WRFRUN_CONTEXT_STATUS = False

        # WRFDA is not necessary
        self.USE_WRFDA: bool = False

        # output directory of ungrib
        self._UNGRIB_OUT_DIR = "./outputs"

        self._WRFRUN_OUTPUT_PATH = ":WRFRUN_OUTPUT_PATH:"
        self._WRFRUN_RESOURCE_PATH = ":WRFRUN_RESOURCE_PATH:"

        self.IS_IN_REPLAY = False

        self.IS_RECORDING = False

        # in this mode, wrfrun will do all things except call the numerical model.
        # all output rules will also not be executed.
        self.FAKE_SIMULATION_MODE = False

    def _get_uri_map(self) -> dict[str, str]:
        """
        Return uri and its value.
        We will use this to register uri when initialize config.

        :return:
        :rtype:
        """
        return {
            self.WRFRUN_TEMP_PATH: self._WRFRUN_TEMP_PATH,
            self.WRFRUN_HOME_PATH: self._WRFRUN_HOME_PATH,
            self.WRFRUN_WORKSPACE_PATH: self._WORK_PATH,
            self.WPS_WORK_PATH: self._WPS_WORK_PATH,
            self.WRF_WORK_PATH: self._WRF_WORK_PATH,
            self.WRFDA_WORK_PATH: self._WRFDA_WORK_PATH,
            self.WRFRUN_REPLAY_WORK_PATH: self._WRFRUN_REPLAY_WORK_PATH,
        }

    @property
    def WRFRUN_REPLAY_WORK_PATH(self) -> str:
        return ":WRFRUN_REPLAY_WORK_PATH:"

    @property
    def WRFRUN_TEMP_PATH(self) -> str:
        """
        Path of the directory storing temporary files.

        :return: Path of the directory storing temporary files.
        :rtype: str
        """
        return ":WRFRUN_TEMP_PATH:"

    @property
    def WRFRUN_HOME_PATH(self) -> str:
        """
        Path of the directory storing wrfrun config files.

        :return: Path of the directory storing temporary files.
        :rtype: str
        """
        return ":WRFRUN_HOME_PATH:"

    @property
    def WRFRUN_WORKSPACE_PATH(self) -> str:
        """
        Path of the wrfrun workspace.

        :return: Path of the wrfrun workspace.
        :rtype: str
        """
        return ":WRFRUN_WORK_PATH:"

    @property
    def WPS_WORK_PATH(self) -> str:
        """
        Path of the directory to run WPS.

        :return: Path of the directory to run WPS.
        :rtype: str
        """
        return ":WRFRUN_WPS_WORK_PATH:"

    @property
    def WRF_WORK_PATH(self) -> str:
        """
        Path of the directory to run WRF.

        :return: Path of the directory to run WRF.
        :rtype: str
        """
        return ":WRFRUN_WRF_WORK_PATH:"

    @property
    def WRFDA_WORK_PATH(self) -> str:
        """
        Path of the directory to run WRFDA.

        :return: Path of the directory to run WRFDA.
        :rtype: str
        """
        return ":WRFRUN_WRFDA_WORK_PATH:"

    @property
    def WRFRUN_WORK_STATUS(self) -> str:
        """
        wrfrun work status.

        :return: wrfrun work status.
        :rtype: str
        """
        return self._WORK_STATUS

    @WRFRUN_WORK_STATUS.setter
    def WRFRUN_WORK_STATUS(self, value: str):
        if not isinstance(value, str):
            value = str(value)
        self._WORK_STATUS = value

    @property
    def UNGRIB_OUT_DIR(self):
        """
        Output directory path of ungrib.
        """
        return self._UNGRIB_OUT_DIR

    @UNGRIB_OUT_DIR.setter
    def UNGRIB_OUT_DIR(self, value: str):
        if not isinstance(value, str):
            value = str(value)
        self._UNGRIB_OUT_DIR = value

    @property
    def WRFRUN_OUTPUT_PATH(self) -> str:
        """
        A marco string represents the save path of output files in wrfrun config.
        """
        return self._WRFRUN_OUTPUT_PATH

    @property
    def WRFRUN_RESOURCE_PATH(self) -> str:
        """
        Return the preserved string used by wrfrun resource files.
        """
        return self._WRFRUN_RESOURCE_PATH

    def check_wrfrun_context(self, error=False) -> bool:
        """
        Check if we're in WRFRun context or not.

        :param error: Raise an error if ``error==True`` when we are not in WRFRun context.
        :type error: bool
        :return: True or False.
        :rtype: bool
        """
        if self._WRFRUN_CONTEXT_STATUS:
            return self._WRFRUN_CONTEXT_STATUS

        if not error:
            logger.warning("You are using wrfrun without entering `WRFRun` context, which may cause some functions don't work.")
            return self._WRFRUN_CONTEXT_STATUS

        logger.error("You need to enter `WRFRun` context to use wrfrun.")
        raise WRFRunContextError("You need to enter `WRFRun` context to use wrfrun.")

    def set_wrfrun_context(self, status: bool):
        """
        Change ``WRFRun`` context status to True or False.

        :param status: ``WRFRun`` context status.
        :type status: bool
        """
        self._WRFRUN_CONTEXT_STATUS = status


class _WRFRunNamelist:
    """
    A class to manage namelist config.
    """
    def __init__(self):
        self._wps_namelist = {}
        self._wrf_namelist = {}
        self._wrfda_namelist = {}
        self._namelist_dict = {}
        self._namelist_id_list = ("param", "geog_static_data", "wps", "wrf", "wrfda")

    def register_custom_namelist_id(self, namelist_id: str) -> bool:
        """
        Register a namelist with a unique id so you can read, update and write it later.

        :param namelist_id: A unique namelist id.
        :type namelist_id: str
        :return: True if register successfully, else False.
        :rtype: bool
        """
        if namelist_id in self._namelist_id_list:
            return False

        else:
            self._namelist_id_list += (namelist_id,)
            return True

    def unregister_custom_namelist_id(self, namelist_id: str):
        """
        Unregister a specified namelist id.
        If unregister successfully, all data about this namelist kind will be lost.

        :param namelist_id: A unique namelist id.
        :type namelist_id: str
        :return:
        :rtype:
        """
        if namelist_id not in self._namelist_id_list:
            return

        self.delete_namelist(namelist_id)
        self._namelist_id_list = tuple(set(self._namelist_id_list) - {namelist_id, })

    def check_namelist_id(self, namelist_id: str) -> bool:
        """
        Check if a namelist id is registered.

        :param namelist_id: Unique namelist ID.
        :type namelist_id: Unique namelist ID.
        :return: True if the ID is registered, else False.
        :rtype: bool
        """
        if namelist_id in self._namelist_id_list:
            return True
        else:
            return False

    def read_namelist(self, file_path: str, namelist_id: str):
        """
        Read namelist.

        :param file_path: Namelist file path.
        :type file_path: str
        :param namelist_id: ``"wps"``, ``"wrf"``, ``"wrfda"``, or any other id you have registered.
        :type namelist_id: str
        """
        # check the file path
        if not exists(file_path):
            logger.error(f"File not found: {file_path}")
            raise FileNotFoundError

        if namelist_id not in self._namelist_id_list:
            logger.error(f"Unknown namelist id: {namelist_id}, register it first.")
            raise ValueError(f"Unknown namelist id: {namelist_id}, register it first.")

        self._namelist_dict[namelist_id] = f90nml.read(file_path).todict()

    def write_namelist(self, save_path: str, namelist_id: str, overwrite=True):
        """
        Write namelist to file.

        :param save_path: Save path.
        :type save_path: str
        :param namelist_id: ``"wps"``, ``"wrf"``, ``"wrfda"``, or any other id you have registered.
        :type namelist_id: str
        :param overwrite: If overwrite the existed file, defaults to ``True``.
        :type overwrite: bool
        """
        if namelist_id not in self._namelist_id_list:
            logger.error(f"Unknown namelist id: {namelist_id}, register it first.")
            raise ValueError(f"Unknown namelist id: {namelist_id}, register it first.")

        if namelist_id not in self._namelist_dict:
            logger.error(f"Can't found custom namelist '{namelist_id}', maybe you forget to read it first")
            raise KeyError(f"Can't found custom namelist '{namelist_id}', maybe you forget to read it first")

        f90nml.Namelist(self._namelist_dict[namelist_id]).write(save_path, force=overwrite)

    def update_namelist(self, new_values: Union[str, dict], namelist_id: str):
        """
        Update value in namelist data.

        You can give your custom namelist file, a whole namelist or a file only contains values you want to change.

        >>> namelist_file = "./namelist.wps"
        >>> WRFRUNConfig.update_namelist(namelist_file, namelist_id="wps")

        >>> namelist_file = "./namelist.wrf"
        >>> WRFRUNConfig.update_namelist(namelist_file, namelist_id="wrf")

        Or give a Dict object contains values you want to change.

        >>> namelist_values = {"ungrib": {"prefix": "./output/FILE"}}
        >>> WRFRUNConfig.update_namelist(namelist_values, namelist_id="wps")

        >>> namelist_values = {"time_control": {"debug_level": 100}}
        >>> WRFRUNConfig.update_namelist(namelist_values, namelist_id="wrf")

        :param new_values: New values.
        :type new_values: str | dict
        :param namelist_id: ``"wps"``, ``"wrf"``, ``"wrfda"``, or any other id you have registered.
        :type namelist_id: str
        """
        if namelist_id not in self._namelist_id_list:
            logger.error(f"Unknown namelist id: {namelist_id}, register it first.")
            raise ValueError(f"Unknown namelist id: {namelist_id}, register it first.")

        elif namelist_id not in self._namelist_dict:
            self._namelist_dict[namelist_id] = new_values
            return

        else:
            reference = self._namelist_dict[namelist_id]

        if isinstance(new_values, str):
            if not exists(new_values):
                logger.error(f"File not found: {new_values}")
                raise FileNotFoundError(f"File not found: {new_values}")
            new_values = f90nml.read(new_values).todict()

        for key in new_values:
            if key in reference:
                reference[key].update(new_values[key])
            else:
                reference[key] = new_values[key]

    def get_namelist(self, namelist_id: str) -> dict:
        """
        Get specific namelist.

        :param namelist_id: ``"wps"``, ``"wrf"``, ``"wrfda"``, or any other id you have registered.
        :type namelist_id: str
        :return: Namelist.
        :rtype: dict
        """
        if namelist_id not in self._namelist_id_list:
            logger.error(f"Unknown namelist id: {namelist_id}, register it first.")
            raise ValueError(f"Unknown namelist id: {namelist_id}, register it first.")
        elif namelist_id not in self._namelist_dict:
            return {}
        else:
            return deepcopy(self._namelist_dict[namelist_id])

    def delete_namelist(self, namelist_id: str):
        """
        Delete specified namelist values.

        :param namelist_id: ``"wps"``, ``"wrf"``, ``"wrfda"``, or any other id you have registered.
        :type namelist_id: str
        :return:
        :rtype:
        """
        if namelist_id not in self._namelist_id_list:
            logger.error(f"Unknown namelist id: {namelist_id}, register it first.")
            raise ValueError(f"Unknown namelist id: {namelist_id}, register it first.")

        if namelist_id not in self._namelist_dict:
            return

        self._namelist_dict.pop(namelist_id)


class WRFRunConfig(_WRFRunConstants, _WRFRunNamelist, _WRFRunResources):
    """
    Class to manage wrfrun config.
    """
    _instance = None
    _initialized = False

    def __init__(self):
        if self._initialized:
            return

        _WRFRunConstants.__init__(self)
        _WRFRunNamelist.__init__(self)
        _WRFRunResources.__init__(self)

        self._config = {}

        # register uri for wrfrun constants
        for key, value in self._get_uri_map().items():
            self.register_resource_uri(key, value)

        self._config_template_file_path = None

        self._initialized = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def set_config_template_path(self, file_path: str):
        """
        Set the path of template file.

        :param file_path: Template file path.
        :type file_path: str
        :return:
        :rtype:
        """
        self._config_template_file_path = file_path

    def load_wrfrun_config(self, config_path: Optional[str] = None):
        """
        Load wrfrun config.

        :param config_path: YAML config file. Defaults to None.
        :type config_path: str
        """
        config_template_path = self.parse_resource_uri(self._config_template_file_path)

        if config_path is not None:
            if not exists(config_path):
                logger.error(f"Config file doesn't exist, copy template config to {config_path}")
                logger.error("Please modify it.")

                if not exists(dirname(config_path)):
                    makedirs(dirname(config_path))

                copyfile(config_template_path, config_path)
                raise FileNotFoundError(config_path)
        else:
            logger.info("Read config template since you doesn't give config file")
            logger.info("A new config file has been saved to './config.toml', you can change and use it latter")

            copyfile(config_template_path, "./config.toml")
            config_path = "./config.toml"

        with open(config_path, "rb") as f:
            self._config = tomli.load(f)

        # register URI for output directory.
        output_path = abspath(self["output_path"])
        self.register_resource_uri(self.WRFRUN_OUTPUT_PATH, output_path)

    def save_wrfrun_config(self, save_path: str):
        """
        Save config to a file.

        :param save_path: Save path of the config.
        :type save_path: str
        """
        save_path = self.parse_resource_uri(save_path)

        path_dir = dirname(save_path)
        if not exists(path_dir):
            makedirs(path_dir)

        with open(save_path, "wb") as f:
            tomli_w.dump(self._config, f)

    def __getitem__(self, item):
        if len(self._config) == 0:
            logger.error("Attempt to read value before load config")
            raise RuntimeError("Attempt to read value before load config")

        return deepcopy(self._config[item])

    def get_input_data_path(self) -> list[str]:
        """
        Return the path of input data.

        :return: Path list.
        :rtype: list
        """
        return deepcopy(self["input_data_path"])

    def get_model_config(self, model_name: str) -> dict:
        """
        Return the config of a NWP model.

        :param model_name: Name of the model. For example, "wrf".
        :type model_name: str
        :return: A dict object.
        :rtype: dict
        """
        if model_name not in self["model"]:
            logger.error(f"Config of model '{model_name}' isn't found in your config file.")
            raise ModelNameError(f"Config of model '{model_name}' isn't found in your config file.")

        return deepcopy(self["model"][model_name])

    def get_log_path(self) -> str:
        """
        Return the save path of logs.

        :return: A directory path.
        :rtype: str
        """
        return self["log_path"]

    def get_socket_server_config(self) -> Tuple[str, int]:
        """
        Return settings of the socket server.

        :return: ("host", port)
        :rtype: tuple
        """
        return self["server_host"], self["server_port"]

    def get_job_scheduler_config(self) -> dict:
        """
        Return the config of job scheduler.

        :return: A dict object.
        :rtype: dict
        """
        return deepcopy(self["job_scheduler"])

    def get_core_num(self) -> int:
        """
        Return the number of CPU cores.
        :return:
        :rtype:
        """
        return self["core_num"]

    def get_ungrib_out_dir_path(self) -> str:
        """
        Get the output directory of ungrib output (WRF intermediate file).

        :return: URI path.
        :rtype: str
        """
        wif_prefix = self.get_namelist("wps")["ungrib"]["prefix"]
        wif_path = f"{self.WPS_WORK_PATH}/{dirname(wif_prefix)}"

        return wif_path

    def get_ungrib_out_prefix(self) -> str:
        """
        Get the prefix string of ungrib output (WRF intermediate file).

        :return: Prefix string of ungrib output (WRF intermediate file).
        :rtype: str
        """
        wif_prefix = self.get_namelist("wps")["ungrib"]["prefix"]
        wif_prefix = basename(wif_prefix)
        return wif_prefix

    def set_ungrib_out_prefix(self, prefix: str):
        """
        Set the prefix string of ungrib output (WRF intermediate file).

        :param prefix: Prefix string of ungrib output (WRF intermediate file).
        :type prefix: str
        :return:
        :rtype:
        """
        self.update_namelist({
            "ungrib": {"prefix": f"{self.UNGRIB_OUT_DIR}/{prefix}"}
        }, "wps")

    def get_metgrid_fg_names(self) -> list[str]:
        """
        Get prefix strings from "fg_name" in namelist "metgrid" section.

        :return: Prefix strings list.
        :rtype: list
        """
        fg_names = self.get_namelist("wps")["metgrid"]["fg_name"]
        fg_names = [basename(x) for x in fg_names]
        return fg_names

    def set_metgrid_fg_names(self, prefix: Union[str, list[str]]):
        """
        Set prefix strings from "fg_name" in namelist "metgrid" section.

        :param prefix: Prefix strings list.
        :type prefix: str | list
        :return:
        :rtype:
        """
        if isinstance(prefix, str):
            prefix = [prefix, ]
        fg_names = [f"{self.UNGRIB_OUT_DIR}/{x}" for x in prefix]
        self.update_namelist({
            "metgrid": {"fg_name": fg_names}
        }, "wps")

    def write_namelist(self, save_path: str, namelist_id: str, overwrite=True):
        save_path = self.parse_resource_uri(save_path)
        super().write_namelist(save_path, namelist_id, overwrite)


WRFRUNConfig = WRFRunConfig()


__all__ = ["WRFRUNConfig"]



from json import loads
from os.path import exists
from shutil import unpack_archive

from .base import ExecutableBase, ExecutableConfig
from .config import WRFRUNConfig
from .error import ExecRegisterError, GetExecClassError
from ..utils import logger

WRFRUN_REPLAY_URI = ":WRFRUN_REPLAY:"


class WRFRunExecutableRegisterCenter:
    """
    Record all executable class.
    """
    _instance = None
    _initialized = False

    def __init__(self):
        if self._initialized:
            return

        self._exec_db = {}

        self._initialized = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def register_exec(self, unique_iq: str, cls: type):
        """
        Register an executable class.

        :param unique_iq: Unique ID.
        :type unique_iq: str
        :param cls: Class.
        :type cls: type
        :return:
        :rtype:
        """
        if unique_iq in self._exec_db:
            logger.error(f"'{unique_iq}' has been registered.")
            raise ExecRegisterError(f"'{unique_iq}' has been registered.")

        self._exec_db[unique_iq] = cls

    def is_registered(self, unique_id: str) -> bool:
        """
        Check if an executable class has been registered.

        :param unique_id: Unique ID.
        :type unique_id: str
        :return: True or False.
        :rtype: bool
        """
        if unique_id in self._exec_db:
            return True
        else:
            return False

    def get_cls(self, unique_id: str) -> type:
        """
        Get a registered executable class.

        :param unique_id: Unique ID.
        :type unique_id: str
        :return: Executable class.
        :rtype: type
        """
        if unique_id not in self._exec_db:
            logger.error(f"Executable class '{unique_id}' not found.")
            raise GetExecClassError(f"Executable class '{unique_id}' not found.")

        return self._exec_db[unique_id]


WRFRUNExecDB = WRFRunExecutableRegisterCenter()


def replay_config_generator(replay_config_file: str):
    """
    Replay the simulations.

    :param replay_config_file: Replay file path.
    :type replay_config_file: str
    :return:
    :rtype:
    """
    logger.info(f"Loading replay resources from: {replay_config_file}")
    work_path = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WRFRUN_REPLAY_WORK_PATH)

    unpack_archive(replay_config_file, work_path, "zip")

    if not exists(f"{work_path}/config.json"):
        logger.error("Can't find replay config in the provided config file.")
        raise FileNotFoundError("Can't find replay config in the provided config file.")

    with open(f"{work_path}/config.json", "r") as f:
        replay_config_list: list[ExecutableConfig] = loads(f.read())

    for _config in replay_config_list:
        args = _config["class_config"]["class_args"]
        kwargs = _config["class_config"]["class_kwargs"]
        executable: ExecutableBase = WRFRUNExecDB.get_cls(_config["name"])(*args, **kwargs)
        executable.load_config(_config)
        yield _config["name"], executable


__all__ = ["WRFRUNExecDB", "replay_config_generator"]



"""
This file contains function to determine schemes for WRF
"""
from dataclasses import dataclass

from wrfrun.utils import logger


@dataclass()
class SchemeLongWave:
    """
    Long wave physics schemes.
    """
    OFF = 0
    RRTM = 1
    CAM = 3
    RRTMG = 4
    NEW_GODDARD = 5
    FLG = 7
    RRTMG_K = 14
    FAST_RRTMG = 24     # for GPU and MIC
    HELD_SUAREZ = 31
    GFDL = 99

    @classmethod
    def get_scheme_id(cls, key: str = "rrtm"):
        """Get the corresponding integer label for long wave radiation scheme

        Args:
            key (str, optional): Name of long wave radiation scheme. Defaults to "rrtm".
        """
        # Here is the map of scheme name and integer label in WRF
        # Reference link: https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/physics.html#longwave-radiation-schemes
        integer_label_map = {
            "off": cls.OFF,
            "rrtm": cls.RRTM,
            "cam": cls.CAM,
            "rrtmg": cls.RRTMG,
            "new-goddard": cls.NEW_GODDARD,
            "flg": cls.FLG,
            "rrtmg-k": cls.RRTMG_K,
            "fast-rrtmg": cls.FAST_RRTMG,
            "held-suarez": cls.HELD_SUAREZ,
            "gfdl": cls.GFDL
        }

        # check if key is in map
        if key not in integer_label_map:
            logger.error(
                f"Key error: {key}. Valid key: {list(integer_label_map.keys())}"
            )
            raise KeyError

        return integer_label_map[key]


@dataclass()
class SchemeShortWave:
    """
    Short wave physics schemes.
    """
    OFF = 0
    DUDHIA = 1
    GODDARD = 2
    CAM = 3
    RRTMG = 4
    NEW_GODDARD = 5
    FLG = 7
    RRTMG_K = 14
    FAST_RRTMG = 24
    EARTH_HELD_SUAREZ_FORCING = 31
    GFDL = 99

    @classmethod
    def get_scheme_id(cls, key: str = "rrtmg"):
        """Get corresponding integer label for short wave radiation scheme

        Args:
            key (str, optional): Name of short wave radiation scheme. Defaults to "rrtmg".
        """
        # Here is the map of scheme name and integer label in WRF
        # Reference link: https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/physics.html#shortwave-radiation-schemes
        integer_label_map = {
            "off": cls.OFF,
            "dudhia": cls.DUDHIA,
            "goddard": cls.GODDARD,
            "cam": cls.CAM,
            "rrtmg": cls.RRTMG,
            "new-goddard": cls.NEW_GODDARD,
            "flg": cls.FLG,
            "rrtmg-k": cls.RRTMG_K,
            "fast-rrtmg": cls.FAST_RRTMG,
            "earth-hs-force": cls.EARTH_HELD_SUAREZ_FORCING,
            "gfdl": cls.GODDARD
        }

        # check if key is in map
        if key not in integer_label_map:
            logger.error(
                f"Key error: {key}. Valid key: {list(integer_label_map.keys())}"
            )
            raise KeyError

        return integer_label_map[key]


@dataclass()
class SchemeCumulus:
    """
    Cumulus parameterization schemes.
    """
    OFF = 0
    KAIN_FRITSCH = 1
    BMJ = 2
    GRELL_FREITAS = 3
    OLD_SAS = 4
    GRELL_3 = 5
    TIEDTKE = 6
    ZHANG_MCFARLANE = 7
    KF_CUP = 10
    MULTI_SCALE_KF = 11
    KIAPS_SAS = 14
    NEW_TIEDTKE = 16
    GRELL_DEVENYI = 93
    NSAS = 96
    OLD_KF = 99

    @classmethod
    def get_scheme_id(cls, key: str = "kf"):
        """Get corresponding integer label for cumulus parameterization scheme

        Args:
            key (str, optional): Name of cumulus parameterization scheme. Defaults to "kf".
        """
        # Here is the map of scheme name and integer label in WRF
        # Reference link: https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/physics.html#cumulus-parameterization
        integer_label_map = {
            "off": cls.OFF,
            "kf": cls.KAIN_FRITSCH,
            "bmj": cls.BMJ,
            "gf": cls.GRELL_FREITAS,
            "old-sas": cls.OLD_SAS,
            "grell-3": cls.GRELL_3,
            "tiedtke": cls.TIEDTKE,
            "zmf": cls.ZHANG_MCFARLANE,
            "kf-cup": cls.KF_CUP,
            "mkf": cls.MULTI_SCALE_KF,
            "kiaps-sas": cls.KIAPS_SAS,
            "nt": cls.NEW_TIEDTKE,
            "gd": cls.GRELL_DEVENYI,
            "nsas": cls.NSAS,
            "old-kf": cls.OLD_KF
        }

        # check if key is in map
        if key not in integer_label_map:
            logger.error(
                f"Key error: {key}. Valid key: {list(integer_label_map.keys())}"
            )
            raise KeyError

        return integer_label_map[key]


@dataclass()
class SchemePBL:
    """
    PBL physics schemes.
    """
    OFF = 0
    YSU = 1
    MYJ = 2
    QNSE_EDMF = 4
    MYNN2 = 5
    MYNN3 = 6
    ACM2 = 7
    BOULAC = 8
    UW = 9
    TEMF = 10
    SHIN_HONG = 11
    GBM = 12
    EEPS = 16
    KEPS = 17
    MRF = 99

    @classmethod
    def get_scheme_id(cls, key: str = "ysu"):
        """Get corresponding integer label for PBL scheme

        Args:
            key (str, optional): Name of PBL scheme. Defaults to "ysu".
        """
        # Here is the map of scheme name and integer label in WRF
        # Reference link: https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/physics.html#pbl-scheme-options
        integer_label_map = {
            "off": cls.OFF,
            "ysu": cls.YSU,
            "myj": cls.MYJ,
            "qe": cls.QNSE_EDMF,
            "mynn2": cls.MYNN2,
            "mynn3": cls.MYNN3,
            "acm2": cls.ACM2,
            "boulac": cls.BOULAC,
            "uw": cls.UW,
            "temf": cls.TEMF,
            "shin-hong": cls.SHIN_HONG,
            "gbm": cls.GBM,
            "eeps": cls.EEPS,
            "keps": cls.KEPS,
            "mrf": cls.MRF
        }

        # check if key is in map
        if key not in integer_label_map:
            logger.error(
                f"Key error: {key}. Valid key: {list(integer_label_map.keys())}"
            )
            raise KeyError

        return integer_label_map[key]


@dataclass()
class SchemeLandSurfaceModel:
    """
    Land surface model physics schemes.
    """
    OFF = 0
    SLAB = 1
    NOAH = 2
    RUC = 3
    NOAH_MP = 4
    CLM4 = 5
    PLEIM_XIU = 7
    SSIB = 8

    @classmethod
    def get_scheme_id(cls, key: str = "noah"):
        """Get corresponding integer label for land surface scheme

        Args:
            key (str, optional): Name of land surface scheme. Defaults to "noah".
        """
        # Here is the map of scheme name and integer label in WRF
        # Reference link: https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/physics.html#lsm-scheme-details-and-references
        integer_label_map = {
            "off": cls.OFF,
            "slab": cls.SLAB,
            "noah": cls.NOAH,
            "ruc": cls.RUC,
            "noah-mp": cls.NOAH_MP,
            "clm4": cls.CLM4,
            "px": cls.PLEIM_XIU,
            "ssib": cls.SSIB
        }

        # check if key is in map
        if key not in integer_label_map:
            logger.error(
                f"Key error: {key}. Valid key: {list(integer_label_map.keys())}"
            )
            raise KeyError

        return integer_label_map[key]


@dataclass()
class SchemeSurfaceLayer:
    """
    Surface layer physics schemes.
    """
    OFF = 0
    MM5 = 1
    MONIN_OBUKHOV = 2
    QNSE = 4
    MYNN = 5
    PLEIM_XIU = 7
    TEMF = 10
    OLD_MM5 = 91

    @classmethod
    def get_scheme_id(cls, key: str = "mm5"):
        """Get corresponding integer label for surface layer scheme

        Args:
            key (str, optional): Name of surface layer scheme. Defaults to "mo".
        """
        # Here is the map of scheme name and integer label in WRF
        # Reference link: https://www2.mmm.ucar.edu/wrf/users/wrf_users_guide/build/html/namelist_variables.html
        integer_label_map = {
            "off": cls.OFF,
            "mm5": cls.MM5,
            "mo": cls.MONIN_OBUKHOV,
            "qnse": cls.QNSE,
            "mynn": cls.MYNN,
            "px": cls.PLEIM_XIU,
            "temf": cls.TEMF,
            "old-mm5": cls.OLD_MM5
        }

        # check if key is in map
        if key not in integer_label_map:
            logger.error(
                f"Key error: {key}. Valid key: {list(integer_label_map.keys())}"
            )
            raise KeyError

        return integer_label_map[key]


__all__ = ["SchemeCumulus", "SchemeLandSurfaceModel", "SchemeLongWave", "SchemePBL", "SchemeShortWave", "SchemeSurfaceLayer"]



from dataclasses import dataclass

from wrfrun.core import WRFRUNConfig


VTABLE_URI = ":WRFRUN_VTABLE:"


@dataclass
class VtableFiles:
    """
    Represent WPS Vtable files (v4.5).
    With the ":/" we can tell from user custom Vtable files.
    """
    AFWAICE = f"{VTABLE_URI}/Vtable.AFWAICE"
    AGRMETSNOW = f"{VTABLE_URI}/Vtable.AGRMETSNOW"
    AGRMETSOIL = f"{VTABLE_URI}/Vtable.AGRMETSOIL"
    AGRMETSOIL2 = f"{VTABLE_URI}/Vtable.AGRMETSOIL2"
    AGRWRF = f"{VTABLE_URI}/Vtable.AGRWRF"
    ARW = f"{VTABLE_URI}/Vtable.ARW.UPP"
    ARWP = f"{VTABLE_URI}/Vtable.ARWp.UPP"
    AVN0P5WRF = f"{VTABLE_URI}/Vtable.AVN0P5WRF"
    AWIP = f"{VTABLE_URI}/Vtable.AWIP"
    CFSR = f"{VTABLE_URI}/Vtable.CFSR"
    CFSR_MEAN = f"{VTABLE_URI}/Vtable.CFSR_mean"
    ECMWF = f"{VTABLE_URI}/Vtable.ECMWF"
    ECMWF_SIGMA = f"{VTABLE_URI}/Vtable.ECMWF_sigma"
    ERA_ML = f"{VTABLE_URI}/Vtable.ERA-interim.ml"
    ERA_PL = f"{VTABLE_URI}/Vtable.ERA-interim.pl"
    GFDL = f"{VTABLE_URI}/Vtable.GFDL"
    GFS = f"{VTABLE_URI}/Vtable.GFS"
    GFS_OZONE = f"{VTABLE_URI}/Vtable.GFS_OZONE"
    GFSENS = f"{VTABLE_URI}/Vtable.GFSENS"
    GODAS = f"{VTABLE_URI}/Vtable.GODAS"
    GSM = f"{VTABLE_URI}/Vtable.GSM"
    ICONM = f"{VTABLE_URI}/Vtable.ICONm"
    ICONP = f"{VTABLE_URI}/Vtable.ICONp"
    JMAGSM = f"{VTABLE_URI}/Vtable.JMAGSM"
    NAM = f"{VTABLE_URI}/Vtable.NAM"
    NARR = f"{VTABLE_URI}/Vtable.NARR"
    NAVY_SST = f"{VTABLE_URI}/Vtable.NavySST"
    NCEP2 = f"{VTABLE_URI}/Vtable.NCEP2"
    NNRP = f"{VTABLE_URI}/Vtable.NNRP"
    NOGAPS = f"{VTABLE_URI}/Vtable.NOGAPS"
    NOGAPS_GFS_SOIL = f"{VTABLE_URI}/Vtable.NOGAPS_needs_GFS_soil"
    RAP_HYBRID_NCEP = f"{VTABLE_URI}/Vtable.RAP.hybrid.ncep"
    RAP_PRESSURE_NCEP = f"{VTABLE_URI}/Vtable.RAP.pressure.ncep"
    RAP_SIGMA_GSD = f"{VTABLE_URI}/Vtable.RAP.sigma.gsd"
    RAPHRRR = f"{VTABLE_URI}/Vtable.raphrrr"
    RUCB = f"{VTABLE_URI}/Vtable.RUCb"
    RUCP = f"{VTABLE_URI}/Vtable.RUCp"
    SREF = f"{VTABLE_URI}/Vtable.SREF"
    SST = f"{VTABLE_URI}/Vtable.SST"
    TCRP = f"{VTABLE_URI}/Vtable.TCRP"
    UKMO_END_GAME = f"{VTABLE_URI}/Vtable.UKMO_ENDGame"
    UKMO_LANDSEA = f"{VTABLE_URI}/Vtable.UKMO_LANDSEA"
    UKMO_NO_HEIGHTS = f"{VTABLE_URI}/Vtable.UKMO_no_heights"


# register uri
if not WRFRUNConfig.check_resource_uri(VTABLE_URI):
    WRFRUNConfig.register_resource_uri(VTABLE_URI, f"{WRFRUNConfig.WPS_WORK_PATH}/ungrib/Variable_Tables")


__all__ = ["VtableFiles"]



from datetime import datetime
from os import makedirs
from os.path import exists, dirname
from typing import Union, List, Tuple

from pandas import date_range
# from seafog import goos_sst_find_data

import cdsapi

from .core.config import WRFRUNConfig
from .utils import logger

CDS_CLIENT = cdsapi.Client()


class ERA5CONFIG:
    """
    A class to store parameters we will use to download ERA5 data from cdsapi.
    """

    # dataset name
    DATASET_ERA5_SINGLE_LEVEL = "reanalysis-era5-single-levels"
    DATASET_ERA5_PRESSURE_LEVEL = "reanalysis-era5-pressure-levels"

    # type name
    TYPE_REANALYSIS = "reanalysis"

    # format name
    FORMAT_NETCDF = "netcdf"
    FORMAT_GRIB = "grib"

    # download format
    DOWNLOAD_ZIP = "zip"
    DOWNLOAD_UNZIP = "unarchived"

    # all level
    PRESSURE_LEVEL = [
        '1', '2', '3',
        '5', '7', '10',
        '20', '30', '50',
        '70', '100', '125',
        '150', '175', '200',
        '225', '250', '300',
        '350', '400', '450',
        '500', '550', '600',
        '650', '700', '750',
        '775', '800', '825',
        '850', '875', '900',
        '925', '950', '975',
        '1000',
    ]

    # variable name
    VARIABLE_2M_TEMPERATURE = "2m_temperature"
    VARIABLE_2M_DEWPOINT_TEMP = "2m_dewpoint_temperature"
    VARIABLE_LANDSEA_MASK = "land_sea_mask"
    VARIABLE_MEAN_SEA_LEVEL_PRESSURE = "mean_sea_level_pressure"
    VARIABLE_SKIN_TEMPERATURE = "skin_temperature"
    VARIABLE_SNOW_DENSITY = "snow_density"
    VARIABLE_SNOW_DEPTH = "snow_depth"
    VARIABLE_SOIL_TEMP_LEVEL_1 = "soil_temperature_level_1"
    VARIABLE_SOIL_TEMP_LEVEL_2 = "soil_temperature_level_2"
    VARIABLE_SOIL_TEMP_LEVEL_3 = "soil_temperature_level_3"
    VARIABLE_SOIL_TEMP_LEVEL_4 = "soil_temperature_level_4"
    VARIABLE_SURFACE_PRESSURE = "surface_pressure"
    VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_1 = "volumetric_soil_water_layer_1"
    VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_2 = "volumetric_soil_water_layer_2"
    VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_3 = "volumetric_soil_water_layer_3"
    VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_4 = "volumetric_soil_water_layer_4"
    VARIABLE_10M_U_WIND = "10m_u_component_of_wind"
    VARIABLE_10M_V_WIND = "10m_v_component_of_wind"
    VARIABLE_U_WIND = "u_component_of_wind"
    VARIABLE_V_WIND = "v_component_of_wind"
    VARIABLE_SPECIFIC_HUMIDITY = "specific_humidity"
    VARIABLE_RELATIVE_HUMIDITY = "relative_humidity"
    VARIABLE_GEOPOTENTIAL = "geopotential"
    VARIABLE_TEMPERATURE = "temperature"

    # name in downloaded data
    NAME_MAP = {
        "2m_temperature": "t2m",
        "10m_u_component_of_wind": "u10",
        "u_component_of_wind": "u",
        "v_component_of_wind": "v",
        "10m_v_component_of_wind": "v10",
        "specific_humidity": "q",
        "geopotential": "z",
        "relative_humidity": "r",
        "temperature": "t",
        "2m_dewpoint_temperature": "d2m",
        "land_sea_mask": "lsm",
        "mean_sea_level_pressure": "msl",
        "skin_temperature": "skt",
        "snow_density": "rsn",
        "snow_depth": "sd",
        "soil_temperature_level_1": "stl1",
        "soil_temperature_level_2": "stl2",
        "soil_temperature_level_3": "stl3",
        "soil_temperature_level_4": "stl4",
        "surface_pressure": "sp",
        "volumetric_soil_water_layer_1": "swvl1",
        "volumetric_soil_water_layer_2": "swvl2",
        "volumetric_soil_water_layer_3": "swvl3",
        "volumetric_soil_water_layer_4": "swvl4"
    }

    # use a dict to distinguish between two types of data
    TYPE_MAP = {
        "reanalysis-era5-single-levels": (
            "2m_temperature",
            "2m_dewpoint_temperature",
            "land_sea_mask",
            "skin_temperature",
            "snow_density",
            "snow_depth",
            "mean_sea_level_pressure",
            "10m_u_component_of_wind",
            "10m_v_component_of_wind",
            "soil_temperature_level_1",
            "soil_temperature_level_2",
            "soil_temperature_level_3",
            "soil_temperature_level_4",
            "surface_pressure",
            "volumetric_soil_water_layer_1",
            "volumetric_soil_water_layer_2",
            "volumetric_soil_water_layer_3",
            "volumetric_soil_water_layer_4"

        ),
        "reanalysis-era5-pressure-levels": (
            "specific_humidity",
            "u_component_of_wind",
            "v_component_of_wind",
            "geopotential",
            "relative_humidity",
            "temperature"
        )
    }


def _check_variables_and_datasets(variables: Union[str, Tuple[str, ...]], dataset: str) -> bool:
    """Check if variables and datasets correspond

    Args:
        variables (str | tuple[str]): Variables type
        dataset (str): Dataset type

    Returns:
        bool: If check passed, return True, else False
    """
    if isinstance(variables, str):
        if variables in ERA5CONFIG.TYPE_MAP[dataset]:
            return True
        else:
            return False
    else:
        for variable in variables:
            if variable not in ERA5CONFIG.TYPE_MAP[dataset]:
                return False

        return True


def _check_pressure_level(pressure_level: Union[str, List[str]]) -> bool:
    """Check pressure level

    Args:
        pressure_level (int | list[int]): A integer value or a list contains pressure values

    Returns:
        bool: If check passed, return True, else False
    """
    valid_pressure_level = [
        '1', '2', '3',
        '5', '7', '10',
        '20', '30', '50',
        '70', '100', '125',
        '150', '175', '200',
        '225', '250', '300',
        '350', '400', '450',
        '500', '550', '600',
        '650', '700', '750',
        '775', '800', '825',
        '850', '875', '900',
        '925', '950', '975',
        '1000',
    ]

    for level in pressure_level:
        if level not in valid_pressure_level:
            return False

    return True


def find_era5_data(date: Union[List[str], List[datetime]], area: Tuple[int, int, int, int], variables: Union[Tuple[str, ...], str], save_path: str,
                   product_type=ERA5CONFIG.TYPE_REANALYSIS, data_format=ERA5CONFIG.FORMAT_NETCDF, dataset=ERA5CONFIG.DATASET_ERA5_SINGLE_LEVEL,
                   download_format=ERA5CONFIG.DOWNLOAD_UNZIP, pressure_level: Union[int, List[int], str, List[str], None] = None, overwrite=False) -> str:
    """
    download era5 data
    Args:
        date: data date, string (for example, `2020-03-25 00:00`) or datetime object, UTC time
        area: range of longitude and latitude, `[lon1, lon2, lat1, lat2]`
        variables: variables, tuple of str or single string
        save_path: save file path
        product_type: product type, default is reanalysis
        data_format: data format, default is netcdf
        dataset: dataset type, default is reanalysis-era5-single-levels
        download_format: zip or unarchived.
        pressure_level: pressure levels.
        overwrite (bool): If the data file exists, force to download it when `overwrite=True`

    Returns: data path

    """
    # check variables and datasets
    if not _check_variables_and_datasets(variables, dataset):
        logger.error(
            f"Variables {variables} and dataset {dataset} doesn't correspond, check it")
        exit(1)

    # check if we need to create directory
    save_folder = dirname(save_path)
    if not exists(save_folder):
        makedirs(save_folder)

    # re-generate area tuple
    area = (area[-1], area[0], area[-2], area[1])

    # parse date
    if isinstance(date[0], str):
        date = [datetime.strptime(_date, "%Y-%m-%d %H:%M")   # type: ignore
                for _date in date]
    year = list(set(_date.strftime("%Y") for _date in date))  # type: ignore
    month = list(set(_date.strftime("%m") for _date in date))  # type: ignore
    day = list(set(_date.strftime("%d") for _date in date))  # type: ignore
    time = list(set(_date.strftime("%H:%M") for _date in date))  # type: ignore

    # sort list
    year.sort()
    month.sort()
    day.sort()
    time.sort()

    # check if it exists
    if exists(save_path) and not overwrite:
        return save_path

    # create params dict
    params_dict = {
        'product_type': product_type,
        'data_format': data_format,
        'download_format': download_format,
        'variable': variables,
        'year': year,
        'month': month,
        'day': day,
        'time': time,
        'area': area,
    }

    # check if we need to add pressure_level to params dict
    if dataset == ERA5CONFIG.DATASET_ERA5_PRESSURE_LEVEL:
        if pressure_level is None:
            logger.error(
                f"You need to provide pressure levels to download data")
            exit(1)
        # convert value to str
        if not isinstance(pressure_level, list):
            pressure_level = [pressure_level]   # type: ignore
        if not isinstance(pressure_level[0], str):  # type: ignore
            pressure_level = [str(int(x))
                              for x in pressure_level]  # type: ignore
        # check
        if _check_pressure_level(pressure_level):   # type: ignore
            params_dict["pressure_level"] = pressure_level
        else:
            logger.error(
                f"You have passed wrong pressure level to download data, check it")
            exit(1)

    # download data
    logger.info(
        f"Downloading data to {save_path}, it may take several tens of minutes, please wait...")
    CDS_CLIENT.retrieve(dataset, params_dict, save_path)

    return save_path


def prepare_wps_input_data(area: Tuple[int, int, int, int]):
    """Download essential data for WPS.

    Args:
        area (Tuple[int, int, int, int]): Range of longitude and latitude, `[lon1, lon2, lat1, lat2]`.
    """
    wrf_config = WRFRUNConfig.get_model_config("wrf")
    # get start and end date from config
    start_date = wrf_config["time"]["start_date"]
    end_date = wrf_config["time"]["end_date"]

    # remove second part
    start_date = start_date[:-3]
    end_date = end_date[:-3]

    # get hour step
    hour_step = wrf_config["time"]["input_data_interval"] // 3600

    # get data save path
    bg_save_path = wrf_config["wps_input_data_folder"]
    sst_save_path = wrf_config["near_goos_data_folder"]

    # download data
    logger.info(f"Download background data of surface level...")
    download_data(start_date, end_date, hour_step, area, f"{bg_save_path}/surface.grib",
                  data_format="grib", data_type="surface", overwrite=True)

    logger.info(f"Download background data of pressure level...")
    download_data(start_date, end_date, hour_step, area, f"{bg_save_path}/pressure.grib",
                  data_format="grib", data_type="pressure", overwrite=True)

    # logger.info(f"Download NearGOOS data...")
    # download_data(start_date, end_date, hour_step, area,
    #               save_path=sst_save_path, data_type="goos", overwrite=True)


def download_data(
        start_date: str,
        end_date: str,
        hour_step: int,
        area: Tuple[int, int, int, int],
        save_path: str,
        data_format="nc",
        data_type="pressure",
        overwrite=False) -> str:
    """Download essential data

    Args:
        start_date (str): Begin date, for example, "2022-05-19 12:00"
        end_date (str): End date, for example, "2022-05-22 18:00"
        hour_step (int): Hour step
        area (tuple): Range of longitude and latitude, `[lon1, lon2, lat1, lat2]`
        save_path (str): Data save path (era5 data) or data save folder path (goos sst)
        data_format (str): Download data format, "nc" or "grib". Default is "nc"
        data_type (str): Download data type, "pressure", "surface" or "goos". Default is "pressure".
        overwrite (bool): If the data file exists, force to download it when `overwrite=True`

    Returns:
        str: Data save path
    """
    # generate date list
    date_list = date_range(
        start_date, end_date, freq=f"{hour_step}H"
    ).strftime("%Y-%m-%d %H:%M").to_list()

    # check format
    if data_format == "nc":
        data_format = ERA5CONFIG.FORMAT_NETCDF
    elif data_format == "grib":
        data_format = ERA5CONFIG.FORMAT_GRIB
    else:
        logger.error(f"Wrong data format: {data_format}")
        raise KeyError

    # check data type
    if data_type == "pressure":
        data_type = ERA5CONFIG.DATASET_ERA5_PRESSURE_LEVEL
        variables = (
            ERA5CONFIG.VARIABLE_GEOPOTENTIAL,
            ERA5CONFIG.VARIABLE_RELATIVE_HUMIDITY,
            ERA5CONFIG.VARIABLE_SPECIFIC_HUMIDITY,
            ERA5CONFIG.VARIABLE_TEMPERATURE,
            ERA5CONFIG.VARIABLE_U_WIND,
            ERA5CONFIG.VARIABLE_V_WIND
        )
        pressure_level = ERA5CONFIG.PRESSURE_LEVEL
    elif data_type == "surface":
        data_type = ERA5CONFIG.DATASET_ERA5_SINGLE_LEVEL
        variables = (
            ERA5CONFIG.VARIABLE_SURFACE_PRESSURE,
            ERA5CONFIG.VARIABLE_MEAN_SEA_LEVEL_PRESSURE,
            ERA5CONFIG.VARIABLE_SKIN_TEMPERATURE,
            ERA5CONFIG.VARIABLE_2M_TEMPERATURE,
            ERA5CONFIG.VARIABLE_2M_DEWPOINT_TEMP,
            ERA5CONFIG.VARIABLE_10M_U_WIND,
            ERA5CONFIG.VARIABLE_10M_V_WIND,
            ERA5CONFIG.VARIABLE_LANDSEA_MASK,
            ERA5CONFIG.VARIABLE_SOIL_TEMP_LEVEL_1,
            ERA5CONFIG.VARIABLE_SOIL_TEMP_LEVEL_2,
            ERA5CONFIG.VARIABLE_SOIL_TEMP_LEVEL_3,
            ERA5CONFIG.VARIABLE_SOIL_TEMP_LEVEL_4,
            ERA5CONFIG.VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_1,
            ERA5CONFIG.VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_2,
            ERA5CONFIG.VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_3,
            ERA5CONFIG.VARIABLE_VOLUMETRIC_SOIL_WATER_LAYER_4,
            ERA5CONFIG.VARIABLE_SNOW_DEPTH,
            ERA5CONFIG.VARIABLE_SNOW_DENSITY
        )
        pressure_level = None
    elif data_type == "goos":
        logger.warning(f"NEAR-GOOS SST data hasn't been supported yet")
        # download sst data
        # for _date in date_list:
        #     _ = goos_sst_find_data(_date, save_path=save_path)

        return ""
    else:
        logger.error(f"Wrong data type: {data_type}")
        raise KeyError

    # download data
    return find_era5_data(date=date_list, area=area, variables=variables,   # type: ignore
                          save_path=save_path, data_format=data_format,
                          dataset=data_type, pressure_level=pressure_level,  # type: ignore
                          download_format=ERA5CONFIG.DOWNLOAD_UNZIP, overwrite=overwrite)


__all__ = ["find_era5_data", "ERA5CONFIG", "download_data", "prepare_wps_input_data"]



from os import listdir
from os.path import exists
from typing import Dict

from xarray import open_dataset

from wrfrun.core import WRFRUNConfig
from wrfrun.utils import logger


def get_metgrid_levels(nc_file: str) -> Dict[str, int]:
    """Read metgrid output file and get metgrid levels, land cat and metgrid soil levels

    Args:
        nc_file (str): Output nc file path

    Returns:
        Dict[str, int]: {num_metgrid_levels: number, num_land_cat: number, num_metgrid_soil_level: number}
    """
    # check file
    if not exists(nc_file):
        logger.error(f"File {nc_file} not exists")
        raise FileNotFoundError

    # read data
    dataset = open_dataset(nc_file)

    # read variables
    num_metgrid_levels = dataset["num_metgrid_levels"].size
    num_land_cat = dataset.attrs["NUM_LAND_CAT"]
    num_metgrid_soil_levels = dataset.attrs["NUM_METGRID_SOIL_LEVELS"]

    return dict(
        num_metgrid_levels=num_metgrid_levels,
        num_land_cat=num_land_cat,
        num_metgrid_soil_levels=num_metgrid_soil_levels
    )


def reconcile_namelist_metgrid(metgrid_path: str):
    """
    There are some settings in WRF namelist that are affected by metgrid output, for example, ``num_metgrid_levels``.
    Namelist should be checked and modified before be used by WRF.

    :param metgrid_path: The path store output from metgrid.exe.
                         If it is None, the default output path will be used.
    :type metgrid_path: str
    :return:
    :rtype:
    """
    logger.info(f"Checking values in WRF namelist and metgrid output ...")
    metgrid_output_name = [x for x in listdir(metgrid_path) if x.endswith(".nc")]
    metgrid_output_name.sort()
    metgrid_output_name = metgrid_output_name[0]

    metgrid_levels = get_metgrid_levels(f"{metgrid_path}/{metgrid_output_name}")

    update_values = {
        "domains": {
            "num_metgrid_levels": metgrid_levels["num_metgrid_levels"],
            "num_metgrid_soil_levels": metgrid_levels["num_metgrid_soil_levels"],
        },
        "physics": {
            "num_land_cat": metgrid_levels["num_land_cat"]
        }
    }

    WRFRUNConfig.update_namelist(update_values, "wrf")


__all__ = ["get_metgrid_levels", "reconcile_namelist_metgrid"]



"""
``wrfrun.run`` module.
"""

import sys
import threading
from os.path import abspath, dirname, exists
from typing import Optional, Tuple, Union

from .core import ExecConfigRecorder, WRFRUNConfig, WRFRunBasicError, WRFRunServer, WRFRunServerHandler, replay_config_generator, stop_server
from .data import prepare_wps_input_data
from .model import clear_model_logs, plot_domain_area
from .pbs import in_pbs, prepare_pbs_script
from .utils import call_subprocess, logger, logger_add_file_handler
from .workspace import prepare_workspace


def confirm_model_area():
    """
    Ask user to check domain area.

    """
    plot_domain_area()

    if not in_pbs():
        # ask user
        logger.warning(f"Check the domain image, is it right?")
        answer = input("Is it right? [y/N]: ")

        answer = answer.lower()

        if answer not in ["y", "yes"]:
            logger.error(f"Change your domain setting and run again")
            exit(1)


class WRFRun:
    """
    ``WRFRun`` is a context class to use all functions in ``wrfrun`` package.
    """
    _instance = None
    _initialized = False

    def __init__(self, config_file: str, init_workspace=True, start_server=False, pbs_mode=False, prepare_wps_data=False, wps_data_area: Optional[Tuple[int, int, int, int]] = None):
        """
        WRFRun, a context class to achieve some goals before and after running WRF, like save a copy of config file, start and close WRFRunServer.

        :param config_file: ``wrfrun`` config file's path.
        :param init_workspace: If True, clean old files in workspace and re-create it.
        :param start_server: Whether to start WRFRunServer, defaults to True.
        :param pbs_mode: If commit this task to the PBS system, defaults to True.
        :param prepare_wps_data: If True, download input datas for WPS first.
        :param wps_data_area: If ``prepare_wps_data==True``, you need to give the area range of input data so download function can download data from ERA5.
        :return:
        """
        if self._initialized:
            return

        # variables for running WRFRunServer
        self._start_server = start_server
        self._wrfrun_server: Union[WRFRunServer, None] = None
        self._wrfrun_server_thread: Union[threading.Thread, None] = None
        self._ip = ""
        self._port = -1

        self._pbs_mode = pbs_mode
        self._init_workspace = init_workspace
        self._prepare_wps_data = prepare_wps_data
        self._wps_data_area = wps_data_area
        self._entry_file_path = abspath(sys.argv[0])
        self._entry_file_dir_path = dirname(self._entry_file_path)
        self._replay_configs = None

        # make sure we can read the config file, because sometimes the user may run the Python script in a different path.
        abs_config_path = f"{self._entry_file_dir_path}/{config_file}"
        WRFRUNConfig.load_wrfrun_config(abs_config_path)

        self._record_output_path = None
        self._record_include_data = False
        self._WRFRUNReplay = ExecConfigRecorder

        self._initialized = True

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def __enter__(self):
        # check workspace
        check_list = [WRFRUNConfig.WPS_WORK_PATH, WRFRUNConfig.WRF_WORK_PATH, WRFRUNConfig.WRFDA_WORK_PATH]
        check_list = [WRFRUNConfig.parse_resource_uri(x) for x in check_list]
        for _path in check_list:
            if not exists(_path) and not self._init_workspace:
                logger.info(f"Force re-create workspace because it is broken.")
                self._init_workspace = True
                break

        # here is the condition we need to initialize workspace:
        # 1. pbs_mode = True and init_workspace = True, do prepare_workspace before committing the task to the PBS system.
        # 2. pbs_mode = False and init_workspace = True, do prepare_workspace.
        if self._pbs_mode and not in_pbs():
            if self._init_workspace:
                prepare_workspace()

            # ask user before commit the task
            confirm_model_area()

            prepare_pbs_script(self._entry_file_path)

            call_subprocess(["qsub", f"{self._entry_file_dir_path}/run.sh"])
            logger.info(f"Work has been submit to PBS system")
            exit(0)

        elif not self._pbs_mode:
            if self._init_workspace:
                prepare_workspace()

            confirm_model_area()

        # save a copy of config to the output path
        WRFRUNConfig.save_wrfrun_config(f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/config.toml")

        # check if we need to start a server
        if self._start_server:
            # start server
            self._start_wrfrun_server()

        # change status
        WRFRUNConfig.set_wrfrun_context(True)

        logger_add_file_handler(WRFRUNConfig.get_log_path())

        if self._prepare_wps_data:
            if self._wps_data_area is None:
                logger.error(f"If you want wrfrun preparing data, you need to give `wps_data_area`")
                raise ValueError(f"If you want wrfrun preparing data, you need to give `wps_data_area`")
            else:
                prepare_wps_input_data(self._wps_data_area)

        logger.info(r"Enter wrfrun context")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # stop thread if needed
        if self._start_server:
            stop_server(self._ip, self._port)     # type: ignore

        if exc_type is None and WRFRUNConfig.IS_RECORDING:
            self._WRFRUNReplay.export_replay_file()
            self._WRFRUNReplay.clear_records()

        # change status
        WRFRUNConfig.set_wrfrun_context(False)

        clear_model_logs()

        logger.info(r"Exit wrfrun context")

    def _start_wrfrun_server(self):
        """
        Start a WRFRunServer.
        """
        # read ip and port settings from config
        socket_ip, socket_port = WRFRUNConfig.get_socket_server_config()

        # get simulate settings
        start_date = WRFRUNConfig.get_model_config("wrf")["time"]["start_date"]
        end_date = WRFRUNConfig.get_model_config("wrf")["time"]["end_date"]

        start_date = start_date[0] if isinstance(start_date, list) else start_date
        end_date = end_date[0] if isinstance(end_date, list) else end_date

        # calculate simulate seconds
        time_delta = end_date - start_date
        simulate_seconds = time_delta.days * 24 * 60 * 60 + time_delta.seconds

        # init server
        self._wrfrun_server = WRFRunServer(
            start_date, simulate_seconds, (socket_ip, socket_port),
            WRFRunServerHandler
        )

        # get ip and port from instance, because the port may change
        self._ip, self._port = self._wrfrun_server.server_address

        logger.info(f"Start socket server on {self._ip}:{self._port}")

        # start server thread
        self._wrfrun_server_thread = threading.Thread(
            target=self._wrfrun_server.serve_forever
        )
        self._wrfrun_server_thread.daemon = True
        self._wrfrun_server_thread.start()

    def record_simulation(self, output_path: str, include_data=False):
        """
        Change settings, so wrfrun can record simulation and generate a replay file.

        :param output_path: Output file path.
        :type output_path: str
        :param include_data: If includes data.
        :type include_data: bool
        :return:
        :rtype:
        """
        self._record_output_path = output_path
        self._record_include_data = include_data
        self._WRFRUNReplay = self._WRFRUNReplay.reinit(save_path=output_path, include_data=include_data)

    def replay_simulation(self, replay_file: str):
        """
        Replay the simulation without any changes.

        :param replay_file:
        :type replay_file:
        :return:
        :rtype:
        """
        WRFRUNConfig.check_wrfrun_context(True)

        if self._replay_configs is not None:
            del self._replay_configs

        self._replay_configs = replay_config_generator(replay_file)

        WRFRUNConfig.IS_IN_REPLAY = True

        try:
            for _, executable in self._replay_configs:
                executable.replay()

        except WRFRunBasicError:
            logger.error("Failed to replay the simulation")

        WRFRUNConfig.IS_IN_REPLAY = False

    def replay_executables(self, replay_file: str):
        """
        Replay the simulation without any changes.

        :param replay_file:
        :type replay_file:
        :return:
        :rtype:
        """
        WRFRUNConfig.check_wrfrun_context(True)

        if self._replay_configs is not None:
            del self._replay_configs

        self._replay_configs = replay_config_generator(replay_file)

        WRFRUNConfig.IS_IN_REPLAY = True

        for name, executable in self._replay_configs:
            yield name, executable

        WRFRUNConfig.IS_IN_REPLAY = False


__all__ = ["WRFRun"]



from cfgrib.xarray_to_grib import to_grib
from numpy.dtypes import DateTime64DType
from pandas import to_datetime
from xarray import DataArray, Dataset

from wrfrun.utils import logger


def create_sst_grib(data: DataArray, save_path: str):
    """
    Create a dataset and save it to a GRIB file.

    Args:
        data: DataArray data, it should at least contain three dimensions: ``["time", "latitude", "longitude"]``.
        save_path: GRIB file path.

    """
    # check the data's dimensions.
    for _dim in ["time", "longitude", "latitude"]:
        if _dim not in data.dims:
            logger.error(f"Essential dimension '{_dim}' not found in data")
            raise KeyError

    if not isinstance(data["time"].dtype, DateTime64DType):     # type: ignore
        data = data.assign_coords(time=to_datetime(data["time"].data))

    data = data.sortby(data["latitude"], ascending=False)

    longitude_start, longitude_stop = data["longitude"][0].data, data["longitude"][-1].data
    latitude_start, latitude_stop = data["latitude"][0].data, data["latitude"][-1].data
    delta_longitude = data["longitude"][1].data - data["longitude"][0].data
    delta_latitude = data["latitude"][0].data - data["latitude"][1].data
    longitude_length = data["longitude"].size
    latitude_length = data["latitude"].size
    points_number = data.size

    data = data.assign_attrs(**{
        "units": "K",
        "long_name": "Sea surface temperature",
        "standard_name": "Sea surface temperature",
        # The following keys and values will be used in GRIB.
        "GRIB_paramId": 34,
        "GRIB_shortName": "sst",
        "GRIB_units": "K",
        "GRIB_name": "Sea surface temperature",
        "GRIB_stepUnits": 1,
        "GRIB_stepType": "instant",
        "GRIB_gridType": "regular_ll",
        "GRIB_iDirectionIncrementInDegrees": delta_longitude,
        "GRIB_iScanNegatively": 0,
        "GRIB_jDirectionIncrementInDegrees": delta_latitude,
        "GRIB_jScanPositively": 0,
        "GRIB_latitudeOfFirstGridPointInDegrees": latitude_start,
        "GRIB_latitudeOfLastGridPointInDegrees": latitude_stop,
        "GRIB_longitudeOfFirstGridPointInDegrees": longitude_start,
        "GRIB_longitudeOfLastGridPointInDegrees": longitude_stop,
        "GRIB_Ny": latitude_length,
        "GRIB_Nx": longitude_length,
        "GRIB_typeOfLevel": "surface",
        # The following keys and values can't be found at ECMWF websites.
        "GRIB_cfName": "unknown",
        "GRIB_cfVarName": "sst",
        "GRIB_dataType": "an",  # Analysis data, defined at https://codes.ecmwf.int/grib/format/mars/type/
        "GRIB_gridDefinitionDescription": "Latitude/Longitude Grid",
        # "GRIB_missingValue": -9999,
        "GRIB_numberOfPoints": points_number,
        "GRIB_totalNumber": 0,
        "GRIB_uvRelativeToGird": 0
    })

    to_grib(
        Dataset({
            "sst": data
        }, attrs={
            "GRIB_centre": "ecmf",
            "GRIB_edition": 1,
        }), save_path
    )


__all__ = ["create_sst_grib"]



from os import environ
from os.path import exists, dirname, abspath

from .core import WRFRUNConfig
from .res import TASKSYS_PBS_TEMPLATE
from .utils import logger


def prepare_pbs_script(main_file_path: str):
    """Prepare the bash script to be submitted to PBS work system.

    Args:
        main_file_path (str): The path of main Python file.
    """
    # check main file
    if not exists(main_file_path):
        logger.error(f"Wrong path of main Python file: {main_file_path}")
        raise FileNotFoundError

    # get absolute path of parent directory
    dir_path = abspath(dirname(main_file_path))

    # read log path and PBS setting from config
    log_path = WRFRUNConfig.get_log_path()
    PBS_setting = WRFRUNConfig.get_job_scheduler_config()

    # set PBS log path
    stdout_log_path = f"{log_path}/PBS.log"
    stderr_log_path = f"{log_path}/PBS.log"

    # set environment parameter
    env_settings = ''
    if len(PBS_setting["env_settings"]) != 0:
        for key in PBS_setting["env_settings"]:
            env_settings += f"{key}={PBS_setting['env_settings'][key]}\n"

    # set command
    exec_cmd = f"{PBS_setting['python_interpreter']} {main_file_path}"

    # read template and write to file
    pbs_template_path = WRFRUNConfig.parse_resource_uri(TASKSYS_PBS_TEMPLATE)
    with open(f"{dir_path}/run.sh", "w") as f:

        with open(pbs_template_path, "r") as f_template:
            template = f_template.read()

        template = template.format(
            STDOUT_LOG_PATH=stdout_log_path,
            STDERR_LOG_PATH=stderr_log_path,
            NODE_NUM=PBS_setting["node_num"],
            CORE_NUM=PBS_setting["core_num"],
            ENV_SETTINGS=env_settings,
            WORK_PATH=dir_path,
            WORK_COMMAND=exec_cmd
        )

        f.write(template)

    logger.info(
        f"PBS script has been generated and write to file {dir_path}/run.sh. Check it and submit it to PBS system to run wrfrun.")


def get_core_num() -> int:
    """Read core num from config.

    Returns:
        int: Core number.
    """
    return WRFRUNConfig["core_num"]


def in_pbs() -> bool:
    """Check if we're in a PBS work.

    Returns:
        bool: True if we're in, False is we aren't.
    """
    # check if we're a PBS task
    if "PBS_ENVIRONMENT" in environ:
        # we're in a PBS task
        return True
    else:
        return False


__all__ = ["prepare_pbs_script", "get_core_num", "in_pbs"]



from os.path import abspath, dirname

from wrfrun.core import WRFRUNConfig


RES_PATH = abspath(dirname(__file__))
WRFRUNConfig.register_resource_uri(WRFRUNConfig.WRFRUN_RESOURCE_PATH, RES_PATH)

CONFIG_TOML_TEMPLATE = ":WRFRUN_RESOURCE_PATH:/config.toml.template"
EXT_NCL_PLOT_SCRIPT = ":WRFRUN_RESOURCE_PATH:/extension/plotgrids.ncl"
TASKSYS_PBS_TEMPLATE = ":WRFRUN_RESOURCE_PATH:/job_scheduler/pbs.template"
TASKSYS_SLURM_TEMPLATE = ":WRFRUN_RESOURCE_PATH:/job_scheduler/slurm.template"
NAMELIST_WRFDA = ":WRFRUN_RESOURCE_PATH:/namelist/namelist.input.da_wrfvar.template"
NAMELIST_DFI = ":WRFRUN_RESOURCE_PATH:/namelist/namelist.input.dfi.template"
NAMELIST_REAL = ":WRFRUN_RESOURCE_PATH:/namelist/namelist.input.real.template"
NAMELIST_WRF = ":WRFRUN_RESOURCE_PATH:/namelist/namelist.input.wrf.template"
NAMELIST_WPS = ":WRFRUN_RESOURCE_PATH:/namelist/namelist.wps.template"

WRFRUNConfig.set_config_template_path(CONFIG_TOML_TEMPLATE)


__all__ = ["CONFIG_TOML_TEMPLATE", "EXT_NCL_PLOT_SCRIPT", "TASKSYS_PBS_TEMPLATE", "TASKSYS_SLURM_TEMPLATE", "NAMELIST_WRFDA", "NAMELIST_DFI", "NAMELIST_REAL", "NAMELIST_WRF", "NAMELIST_WPS"]



from os import listdir
from os.path import exists
from typing import OrderedDict, Union

import numpy as np
from xarray import DataArray

from wrfrun.core import WRFRUNConfig
from wrfrun.utils import logger

# for example: 00001-00200.00201-00400
DATA_NAME_TEMPLATE = "{}-{}.{}-{}"


def _get_data_type(wordsize: int) -> type:
    """Get data type based on wordsize value in index file.

    Args:
        wordsize (int): Wordsize in index file.

    Returns:
        type: NumPy value type.
    """
    # define map dict
    map_dict = {
        1: np.int8,
        2: np.int16,
        4: np.int32
    }

    return map_dict[wordsize]


def _get_clip_area(index_area: tuple[int, int, int, int], row_num: int, col_num: int, tile_x: int, tile_y: int) -> tuple[int, int, int, int]:
    """Get clip area.

    Args:
        index_area (tuple[int, int, int, int]): Full area index.
        row_num (int): Row number of the tile.
        col_num (int): Column number of the file.
        tile_x (int): X size of the tile.
        tile_y (int): Y size of the tile.

    Returns:
        tuple[int, int, int, int]: Clip area.
    """
    # calculate tile area
    tile_area = (
        col_num * tile_x + 1,
        col_num * tile_x + tile_x,
        row_num * tile_y + 1,
        row_num * tile_y + tile_y,
    )

    # generate clip area
    clip_area = (
        0 if index_area[0] <= tile_area[0] else index_area[0] % tile_x - 1,
        tile_x if index_area[1] >= tile_area[1] else index_area[1] % tile_x - 1,
        0 if index_area[2] <= tile_area[2] else index_area[2] % tile_y - 1,
        tile_y if index_area[3] >= tile_area[3] else index_area[3] % tile_y - 1,
    )

    return clip_area


def parse_geographical_data_index(index_path: str) -> OrderedDict:
    """Read geographical data index file.

    Args:
        index_path (str): Index file path.

    Returns:
        dict: Info stored in dict.
    """
    # since the index file is very similar to fortran namelist file,
    # we can manually add "&index" and "/" and parse it as a namelist
    # temp file store path
    temp_file = f"{WRFRUNConfig.WRFRUN_TEMP_PATH}/geogrid_data.index"
    temp_file = WRFRUNConfig.parse_resource_uri(temp_file)

    # open file and add header and tail
    with open(index_path, "r") as _index_file:
        with open(temp_file, "w") as _temp_file:

            _temp_file.write("&index\n")
            _temp_file.write(_index_file.read())
            _temp_file.write("/")

    # read index
    WRFRUNConfig.read_namelist(temp_file, "geog_static_data")

    return WRFRUNConfig.get_namelist("geog_static_data")["index"]


def parse_geographical_data_file(file_path: str, wordsize: int, endian: str, tile_shape: tuple[int, ...],
                                 area: Union[tuple[int, ...], None] = None, miss_value: Union[int, float, None] = None) -> np.ndarray:
    """Read geographical data file.

    Args:
        file_path (str): File path.
        wordsize (int): How many bytes are used to store value in data file.
        endian (str): "big" or "little".
        tile_shape (tuple[int, ...]): The raw shape of the tile. Can be 2D or 3D.
        area (Union[tuple[int, ...], None], optional): The range (x_start, x_stop, y_start, y_stop, ...) of data you want to read. Defaults to None.
        miss_value (Union[int, float, None], optional): The value which represents NaN. Defaults to None.

    Returns:
        np.ndarray: NumPy array object.
    """
    # get data type
    data_type = _get_data_type(wordsize)

    # read data
    data = np.fromfile(file_path, dtype=data_type)

    # swap byte if need
    if endian == "big":
        data = data.byteswap()

    # reshape
    data = data.reshape(tile_shape)

    # clip
    if area:
        # check area
        if len(area) % 2 != 0:
            logger.error(
                f"The length of `area` must be even, but is {len(area)}")
            exit(1)

        area_array = np.asarray(area).reshape(-1, 2)
        slice_index = tuple((slice(i[0], i[1]) for i in area_array))

        if len(slice_index) == 2:
            slice_index += (slice(None), )

        data = data[slice_index[::-1]]    # type: ignore

    # fill nan
    if miss_value:
        data[data == miss_value] = np.nan

    return data


def read_geographical_static_data(geog_data_folder_path: str, name: str, area: Union[tuple[float, float, float, float], None] = None) -> DataArray:
    """Read WPS geographical static data

    Args:
        geog_data_folder_path (str): Data folder path.
        name (str): Name that will be used to create DataArray.
        area (Union[tuple[float, float, float, float], None]): Longitude and latitude area (lon_start, lon_stop, lat_start, lat_stop). Defaults to None.

    Returns:
        DataArray: DataArray object.
    """
    # check if folder exists
    if not exists(geog_data_folder_path):
        logger.error(f"Can't find folder {geog_data_folder_path}")
        exit(1)

    # parse index file first
    index_path = f"{geog_data_folder_path}/index"
    index_data = parse_geographical_data_index(index_path)

    # extract info to read data
    # # check essential key
    if "wordsize" not in index_data:
        logger.error(
            f"Can't find key `wordsize` in index file, maybe it is corrupted.")
        exit(1)
    # # extract info
    wordsize = index_data["wordsize"]
    endian = "little" if "endian" not in index_data else index_data["endian"]
    miss_value = None if "missing_value" not in index_data else index_data["missing_value"]
    tile_shape = []
    for key in ["tile_z", "tile_y", "tile_x"]:
        if key in index_data:
            tile_shape.append(index_data[key])
    tile_shape = tuple(tile_shape)
    known_lat = index_data["known_lat"]
    known_lon = index_data["known_lon"]
    dx = index_data["dx"]
    dy = index_data["dy"]

    # calculate area
    if area:
        # read resolution, latitude and longitude
        index_area = (
            int((area[0] - known_lon) // dx),
            int((area[1] - known_lon) // dx),
            int((area[2] - known_lat) // dy),
            int((area[3] - known_lat) // dy),
        )
        # check if negative value exists
        if (
            index_area[0] < 0 or
            index_area[2] < 0
        ):
            logger.warning(f"Part of your area has exceeded data's area")
            # set negative value to 0
            index_area = tuple((i if i >= 0 else 0 for i in index_area))
    else:
        logger.warning(f"You want to read all data, which may be very large")
        index_area = None

    # find the file we need to read
    if index_area:
        # # calculate tile index number
        tile_index_num = (
            index_area[0] // tile_shape[-1],
            index_area[1] // tile_shape[-1],
            index_area[2] // tile_shape[-2],
            index_area[3] // tile_shape[-2],
        )

        filenames = []
        # # generate filenames and clip area
        for row_num in range(tile_index_num[2], tile_index_num[3] + 1):

            _names = []
            for col_num in range(tile_index_num[0], tile_index_num[1] + 1):

                _names.append(
                    [
                        DATA_NAME_TEMPLATE.format(
                            str(col_num * tile_shape[-1] + 1).rjust(5, '0'),
                            str((col_num + 1) * tile_shape[-1]).rjust(5, '0'),
                            str(row_num * tile_shape[-2] + 1).rjust(5, '0'),
                            str((row_num + 1) * tile_shape[-2]).rjust(5, '0'),
                        ), _get_clip_area(index_area, row_num, col_num, tile_shape[-1], tile_shape[-2])   # type: ignore
                    ]
                )

            filenames.append(_names)
    else:
        raw_filenames = [x for x in listdir(
            geog_data_folder_path) if x != "index"]
        raw_filenames.sort()

        # parse the last file to get row number and column number
        _last_filename = raw_filenames[-1]
        total_col_num = int(_last_filename.split(
            ".")[0].split("-")[1]) // tile_shape[-1]
        total_row_num = int(_last_filename.split(
            ".")[1].split("-")[1]) // tile_shape[-2]

        filenames = []

        for row_num in range(total_row_num):

            _names = []
            for col_num in range(total_col_num):

                _names.append(
                    [
                        raw_filenames[row_num * total_col_num + col_num], None
                    ]
                )

            filenames.append(_names)

    # read and concatenate
    array = []
    for _row in filenames:

        _array = []
        for _col in _row:

            _array.append(parse_geographical_data_file(
                f"{geog_data_folder_path}/{_col[0]}", wordsize, endian, tile_shape, _col[1], miss_value))

        # concatenate _array
        array.append(np.concatenate(_array, axis=-1))

    array = np.concatenate(array, axis=-2)

    # get the longitude and latitude of the start point
    if index_area:
        longitude = known_lon + dx * index_area[0]
        latitude = known_lat + dy * index_area[2]
    else:
        longitude = known_lon
        latitude = known_lat

    longitude = np.arange(array.shape[-1]) * dx + longitude
    latitude = np.arange(array.shape[-2]) * dy + latitude
    levels = np.arange(array.shape[-3])

    return DataArray(
        name=name, data=array,
        dims=["levels", "latitude", "longitude"],
        coords={
            "longitude": longitude,
            "latitude": latitude,
            "levels": levels
        }, attrs=index_data
    )


__all__ = ["parse_geographical_data_index", "parse_geographical_data_file", "read_geographical_static_data"]



from os import listdir
from os.path import abspath, basename, exists
from shutil import copyfile, move, rmtree
from typing import Optional

from wrfrun.core import ExecutableBase, FileConfigDict, InputFileError, NamelistError, WRFRUNConfig, WRFRUNExecDB
from wrfrun.utils import logger
from ._metgrid import reconcile_namelist_metgrid
from ._ndown import process_after_ndown
from .namelist import prepare_dfi_namelist, prepare_wps_namelist, prepare_wrf_namelist, prepare_wrfda_namelist
from .vtable import VtableFiles
from ..base import NamelistName


def _check_namelist_preparation():
    if len(WRFRUNConfig.get_namelist("wps")) == 0:
        prepare_wps_namelist()

    if len(WRFRUNConfig.get_namelist("wrf")) == 0:
        prepare_wrf_namelist()

    if len(WRFRUNConfig.get_namelist("wrfda")) == 0:
        prepare_wrfda_namelist()


class GeoGrid(ExecutableBase):
    """
    Execute "geogrid.exe".
    """

    def __init__(self, geogrid_tbl_file: Optional[str] = None, core_num: Optional[int] = None):
        """
        Execute "geogrid.exe".

        :param geogrid_tbl_file: Custom GEOGRID.TBL file path. Defaults to None.
        :type geogrid_tbl_file: str
        :param core_num: An positive integer number of used core numbers. ``mpirun`` will be used to execute geogrid.exe if ``core_num != None``.
        :type core_num: int
        """
        if isinstance(core_num, int) and core_num <= 0:
            logger.warning(f"`core_num` should be greater than 0")
            core_num = None

        if core_num is None:
            mpi_use = False
            mpi_cmd = None
            mpi_core_num = core_num

        else:
            mpi_use = True
            mpi_cmd = "mpirun"
            mpi_core_num = core_num

        super().__init__(name="geogrid", cmd="./geogrid.exe", work_path=WRFRUNConfig.WPS_WORK_PATH, mpi_use=mpi_use, mpi_cmd=mpi_cmd, mpi_core_num=mpi_core_num)

        self.geogrid_tbl_file = geogrid_tbl_file

        _check_namelist_preparation()

    def generate_custom_config(self):
        """
        Get and store namelist.
        :return:
        :rtype:
        """
        self.custom_config.update(
            {
                "namelist": WRFRUNConfig.get_namelist("wps"),
                "geogrid_tbl_file": self.geogrid_tbl_file
            }
        )

    def load_custom_config(self):
        WRFRUNConfig.update_namelist(self.custom_config["namelist"], "wps")
        self.geogrid_tbl_file = self.custom_config["geogrid_tbl_file"]

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        WRFRUNConfig.WRFRUN_WORK_STATUS = "geogrid"

        if not WRFRUNConfig.IS_IN_REPLAY:
            if self.geogrid_tbl_file is not None:
                tbl_file: FileConfigDict = {
                    "file_path": self.geogrid_tbl_file, "save_path": f"{WRFRUNConfig.WPS_WORK_PATH}/geogrid",
                    "save_name": "GEOGRID.TBL", "is_data": False, "is_output": False
                }
                self.add_input_files(tbl_file)

        super().before_exec()

        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WPS_WORK_PATH}/{NamelistName.WPS}", "wps")

    def after_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:
            self.add_output_files(save_path=self._log_save_path, startswith="geogrid.log", outputs=NamelistName.WPS)
            self.add_output_files(save_path=self._output_save_path, startswith="geo_em")

        super().after_exec()

        logger.info(f"All geogrid output files have been copied to {WRFRUNConfig.parse_resource_uri(self._output_save_path)}")


class LinkGrib(ExecutableBase):
    """
    Run command: "./link_grib.csh".
    """

    def __init__(self, grib_dir_path: str):
        """
        Execute "link_grib.csh".

        :param grib_dir_path: GRIB data path. Absolute path is recommended.
        :type grib_dir_path: str
        """
        self._link_grib_input_path = "./input_grib_data_dir"

        super().__init__(name="link_grib", cmd=["./link_grib.csh", f"{self._link_grib_input_path}/*", "."], work_path=WRFRUNConfig.WPS_WORK_PATH)
        self.grib_dir_path = grib_dir_path

    def generate_custom_config(self):
        self.class_config["class_args"] = (self.grib_dir_path,)

    def replay(self):
        self()

    def before_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:

            logger.debug(f"Input data are: {self.grib_dir_path}")
            _grib_dir_path = abspath(self.grib_dir_path)

            if not exists(_grib_dir_path):
                logger.error(f"GRIB file directory not found: {_grib_dir_path}")
                raise FileNotFoundError(f"GRIB file directory not found: {_grib_dir_path}")

            save_path = f"{WRFRUNConfig.WPS_WORK_PATH}/{self._link_grib_input_path}"
            save_path = WRFRUNConfig.parse_resource_uri(save_path)
            if exists(save_path):
                rmtree(save_path)

            for _file in listdir(_grib_dir_path):
                _file_config: FileConfigDict = {
                    "file_path": f"{_grib_dir_path}/{_file}",
                    "save_path": f"{WRFRUNConfig.WPS_WORK_PATH}/{self._link_grib_input_path}",
                    "save_name": _file, "is_data": True, "is_output": False,
                }
                self.add_input_files(_file_config)

        super().before_exec()


class UnGrib(ExecutableBase):
    """
    Execute "ungrib.exe".
    """

    def __init__(self, vtable_file: Optional[str] = None, input_data_path: Optional[str] = None):
        """
        Execute "ungrib.exe".

        :param vtable_file: Path of the Vtable file "ungrib.exe" used.
        :type vtable_file: str
        :param input_data_path: Path of the directory stores input GRIB2 files.
        :type input_data_path: str
        """
        super().__init__(name="ungrib", cmd="./ungrib.exe", work_path=WRFRUNConfig.WPS_WORK_PATH)

        self.vtable_file = vtable_file
        self.input_data_path = input_data_path

        _check_namelist_preparation()

    def call_link_grib(self):
        """
        Execute "link_grib.csh" if needed.

        :return:
        :rtype:
        """
        if self.input_data_path is None:
            self.input_data_path = WRFRUNConfig.get_input_data_path()

        LinkGrib(self.input_data_path)()

    def generate_custom_config(self):
        """
        Get and store namelist.

        :return:
        :rtype:
        """
        self.custom_config.update(
            {
                "namelist": WRFRUNConfig.get_namelist("wps"),
                "vtable_file": self.vtable_file
            }
        )

    def load_custom_config(self):
        self.vtable_file = self.custom_config["vtable_file"]
        WRFRUNConfig.update_namelist(self.custom_config["namelist"], "wps")

    def replay(self):
        super().__call__()

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        WRFRUNConfig.WRFRUN_WORK_STATUS = "ungrib"

        if not WRFRUNConfig.IS_IN_REPLAY:
            if self.vtable_file is None:
                self.vtable_file = VtableFiles.ERA_PL

            _file_config: FileConfigDict = {
                "file_path": self.vtable_file,
                "save_path": WRFRUNConfig.WPS_WORK_PATH,
                "save_name": "Vtable",
                "is_data": False,
                "is_output": False
            }
            self.add_input_files(_file_config)

        super().before_exec()

        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WPS_WORK_PATH}/{NamelistName.WPS}", "wps")

    def after_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:
            self.add_output_files(output_dir=WRFRUNConfig.get_ungrib_out_dir_path(), save_path=self._output_save_path, startswith=WRFRUNConfig.get_ungrib_out_prefix())
            self.add_output_files(save_path=self._log_save_path, outputs=["ungrib.log", "namelist.wps"])

        super().after_exec()

        logger.info(f"All ungrib output files have been copied to {WRFRUNConfig.parse_resource_uri(self._output_save_path)}")

    def __call__(self):
        self.call_link_grib()

        super().__call__()


class MetGrid(ExecutableBase):
    """
    Execute "metgrid.exe".
    """

    def __init__(self, geogrid_data_path: Optional[str] = None, ungrib_data_path: Optional[str] = None, core_num: Optional[int] = None):
        """
        Execute "metgrid.exe".

        :param geogrid_data_path: Directory path of outputs from geogrid.exe. If None, tries to use the output path specified by config file.
        :type geogrid_data_path: str
        :param ungrib_data_path: Directory path of outputs from ungrib.exe. If None, tries to use the output path specified by config file.
        :type ungrib_data_path: str
        :param core_num: An positive integer number of used core numbers. ``mpirun`` will be used to execute geogrid.exe if ``core_num != None``.
        :type core_num: int
        """
        if isinstance(core_num, int) and core_num <= 0:
            logger.warning(f"`core_num` should be greater than 0")
            core_num = None

        if core_num is None:
            mpi_use = False
            mpi_cmd = None
            mpi_core_num = core_num

        else:
            mpi_use = True
            mpi_cmd = "mpirun"
            mpi_core_num = core_num

        super().__init__(name="metgrid", cmd="./metgrid.exe", work_path=WRFRUNConfig.WPS_WORK_PATH, mpi_use=mpi_use, mpi_cmd=mpi_cmd, mpi_core_num=mpi_core_num)

        self.geogrid_data_path = geogrid_data_path
        self.ungrib_data_path = ungrib_data_path

        _check_namelist_preparation()

    def generate_custom_config(self):
        """
        Get and store namelist.
        :return:
        :rtype:
        """
        self.custom_config.update(
            {
                "geogrid_data_path": self.geogrid_data_path,
                "ungrib_data_path": self.ungrib_data_path,
                "namelist": WRFRUNConfig.get_namelist("wps"),
            }
        )

    def load_custom_config(self):
        self.geogrid_data_path = self.custom_config["geogrid_data_path"]
        self.ungrib_data_path = self.custom_config["ungrib_data_path"]
        WRFRUNConfig.update_namelist(self.custom_config["namelist"], "wps")

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        WRFRUNConfig.WRFRUN_WORK_STATUS = "metgrid"

        if not WRFRUNConfig.IS_IN_REPLAY and not WRFRUNConfig.FAKE_SIMULATION_MODE:
            # check input of metgrid.exe
            # try to search input files in the output path if workspace is clear.
            file_list = listdir(WRFRUNConfig.parse_resource_uri(WRFRUNConfig.WPS_WORK_PATH))

            if "geo_em.d01.nc" not in file_list:

                if self.geogrid_data_path is None:
                    self.geogrid_data_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/geogrid"
                geogrid_data_path = WRFRUNConfig.parse_resource_uri(self.geogrid_data_path)

                if not exists(geogrid_data_path) or "geo_em.d01.nc" not in listdir(geogrid_data_path):
                    logger.error(f"Can't find geogrid outputs in WPS_WORK_PATH and your outputs directory, which is essential to run metgrid")
                    raise FileNotFoundError(f"Can't find geogrid outputs in WPS_WORK_PATH and your outputs directory, which is essential to run metgrid")

                else:
                    geogrid_file_list = [x for x in listdir(geogrid_data_path) if x.startswith("geo_em.d")]
                    for _file in geogrid_file_list:
                        _file_config = {
                            "file_path": f"{self.geogrid_data_path}/{_file}",
                            "save_path": WRFRUNConfig.WPS_WORK_PATH,
                            "save_name": _file,
                            "is_data": True,
                            "is_output": True
                        }
                        self.add_input_files(_file_config)

            ungrib_output_dir = WRFRUNConfig.parse_resource_uri(WRFRUNConfig.get_ungrib_out_dir_path())
            if basename(ungrib_output_dir) not in file_list or len(listdir(ungrib_output_dir)) == 0:

                if self.ungrib_data_path is None:
                    self.ungrib_data_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/ungrib"

                ungrib_data_path = WRFRUNConfig.parse_resource_uri(self.ungrib_data_path)

                if not exists(ungrib_data_path) or len(listdir(ungrib_data_path)) == 0:
                    logger.error(f"Can't find ungrib outputs in WPS_WORK_PATH and your outputs directory, which is essential to run metgrid")
                    raise FileNotFoundError(f"Can't find ungrib outputs in WPS_WORK_PATH and your outputs directory, which is essential to run metgrid")

                else:
                    ungrib_file_list = [x for x in listdir(ungrib_data_path)]
                    for _file in ungrib_file_list:
                        _file_config: FileConfigDict = {
                            "file_path": f"{self.ungrib_data_path}/{_file}",
                            "save_path": WRFRUNConfig.get_ungrib_out_dir_path(),
                            "save_name": _file,
                            "is_data": True,
                            "is_output": True
                        }
                        self.add_input_files(_file_config)

        super().before_exec()

        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WPS_WORK_PATH}/{NamelistName.WPS}", "wps")

    def after_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:
            self.add_output_files(save_path=self._log_save_path, startswith="metgrid.log", outputs="namelist.wps")
            self.add_output_files(save_path=self._output_save_path, startswith="met_em")

        super().after_exec()

        logger.info(f"All metgrid output files have been copied to {WRFRUNConfig.parse_resource_uri(self._output_save_path)}")


class Real(ExecutableBase):
    """
    Execute "real.exe".
    """

    def __init__(self, metgrid_data_path: Optional[str] = None, core_num: Optional[int] = None):
        """
        Execute "real.exe".

        :param metgrid_data_path: Path of the directory stores outputs from "metgrid.exe".
        :type metgrid_data_path: str
        :param core_num: An positive integer number of used core numbers. ``mpirun`` will be used to execute geogrid.exe if ``core_num != None``.
        :type core_num: int
        """
        if isinstance(core_num, int) and core_num <= 0:
            logger.warning(f"`core_num` should be greater than 0")
            core_num = None

        if core_num is None:
            mpi_use = False
            mpi_cmd = None
            mpi_core_num = core_num

        else:
            mpi_use = True
            mpi_cmd = "mpirun"
            mpi_core_num = core_num

        _check_namelist_preparation()

        super().__init__(name="real", cmd="./real.exe", work_path=WRFRUNConfig.WRF_WORK_PATH, mpi_use=mpi_use, mpi_cmd=mpi_cmd, mpi_core_num=mpi_core_num)

        self.metgrid_data_path = metgrid_data_path

    def generate_custom_config(self):
        self.custom_config["metgrid_data_path"] = self.metgrid_data_path
        self.custom_config.update(
            {
                "namelist": WRFRUNConfig.get_namelist("wrf"),
                "metgrid_data_path": self.metgrid_data_path
            }
        )

    def load_custom_config(self):
        self.metgrid_data_path = self.custom_config["metgrid_data_path"]
        WRFRUNConfig.update_namelist(self.custom_config["namelist"], "wrf")

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        WRFRUNConfig.WRFRUN_WORK_STATUS = "real"

        if not WRFRUNConfig.IS_IN_REPLAY and not WRFRUNConfig.FAKE_SIMULATION_MODE:
            if self.metgrid_data_path is None:
                self.metgrid_data_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/metgrid"

            metgrid_data_path = WRFRUNConfig.parse_resource_uri(self.metgrid_data_path)
            reconcile_namelist_metgrid(metgrid_data_path)

            file_list = [x for x in listdir(metgrid_data_path) if x.startswith("met_em")]
            for _file in file_list:
                _file_config: FileConfigDict = {
                    "file_path": f"{self.metgrid_data_path}/{_file}",
                    "save_path": WRFRUNConfig.WRF_WORK_PATH,
                    "save_name": _file,
                    "is_data": True,
                    "is_output": True
                }
                self.add_input_files(_file_config)

        super().before_exec()

        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WRF_WORK_PATH}/{NamelistName.WRF}", "wrf")

    def after_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:
            self.add_output_files(save_path=self._output_save_path, startswith=("wrfbdy", "wrfinput", "wrflow"))
            self.add_output_files(save_path=self._log_save_path, startswith="rsl.", outputs="namelist.input")

        super().after_exec()

        logger.info(f"All real output files have been copied to {WRFRUNConfig.parse_resource_uri(self._output_save_path)}")


class WRF(ExecutableBase):
    """
    Execute "wrf.exe".
    """

    def __init__(self, input_file_dir_path: Optional[str] = None, restart_file_dir_path: Optional[str] = None, save_restarts=False, core_num: Optional[int] = None):
        """
        Execute "wrf.exe"

        :param input_file_dir_path: Path of the directory that stores input data for "wrf.exe".
        :type input_file_dir_path: str
        :param restart_file_dir_path: Path of the directory that stores restart files for "wrf.exe".
        :type restart_file_dir_path: str
        :param save_restarts: If saving restart files from "wrf.exe". Defaults to False.
        :type save_restarts: bool
        :param core_num: An positive integer number of used core numbers. ``mpirun`` will be used to execute geogrid.exe if ``core_num != None``.
        :type core_num: int
        """
        if isinstance(core_num, int) and core_num <= 0:
            logger.warning(f"`core_num` should be greater than 0")
            core_num = None

        if core_num is None:
            mpi_use = False
            mpi_cmd = None
            mpi_core_num = core_num

        else:
            mpi_use = True
            mpi_cmd = "mpirun"
            mpi_core_num = core_num

        _check_namelist_preparation()

        super().__init__(name="wrf", cmd="./wrf.exe", work_path=WRFRUNConfig.WRF_WORK_PATH, mpi_use=mpi_use, mpi_cmd=mpi_cmd, mpi_core_num=mpi_core_num)

        self.input_file_dir_path = input_file_dir_path
        self.restart_file_dir_path = restart_file_dir_path
        self.save_restarts = save_restarts

    def generate_custom_config(self):
        self.custom_config.update(
            {
                "input_file_dir_path": self.input_file_dir_path,
                "restart_file_dir_path": self.restart_file_dir_path,
                "namelist": WRFRUNConfig.get_namelist("wrf")
            }
        )

    def load_custom_config(self):
        self.input_file_dir_path = self.custom_config["input_file_dir_path"]
        self.restart_file_dir_path = self.custom_config["restart_file_dir_path"]
        WRFRUNConfig.update_namelist(self.custom_config["namelist"], "wrf")

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        # help wrfrun to make sure the input file is from real or ndown.
        last_work_status = WRFRUNConfig.WRFRUN_WORK_STATUS
        if last_work_status not in ["real", "ndown"]:
            last_work_status = ""
        WRFRUNConfig.WRFRUN_WORK_STATUS = "wrf"

        if not WRFRUNConfig.IS_IN_REPLAY and not WRFRUNConfig.FAKE_SIMULATION_MODE:
            if self.input_file_dir_path is None:
                if last_work_status == "":
                    # assume we already have outputs from real.exe.
                    self.input_file_dir_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/real"
                    is_output = False
                else:
                    self.input_file_dir_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/{last_work_status}"
                    is_output = True

            else:
                is_output = False

            input_file_dir_path = WRFRUNConfig.parse_resource_uri(self.input_file_dir_path)

            if exists(input_file_dir_path):
                file_list = [x for x in listdir(input_file_dir_path) if x != "logs"]

                for _file in file_list:
                    _file_config: FileConfigDict = {
                        "file_path": f"{self.input_file_dir_path}/{_file}",
                        "save_path": WRFRUNConfig.WRF_WORK_PATH,
                        "save_name": _file,
                        "is_data": True,
                        "is_output": is_output
                    }
                    self.add_input_files(_file_config)

            if WRFRUNConfig.get_model_config("wrf")["restart_mode"]:
                if self.restart_file_dir_path is None:
                    logger.error("You need to specify the restart file if you want to restart WRF.")
                    raise InputFileError("You need to specify the restart file if you want to restart WRF.")

                restart_file_dir_path = WRFRUNConfig.parse_resource_uri(self.restart_file_dir_path)

                if not exists(restart_file_dir_path):
                    logger.error(f"Restart files not found: {restart_file_dir_path}")
                    raise FileNotFoundError(f"Restart files not found: {restart_file_dir_path}")

                file_list = [x for x in listdir(restart_file_dir_path) if x.startswith("wrfrst")]
                for _file in file_list:
                    _file_config: FileConfigDict = {
                        "file_path": f"{self.restart_file_dir_path}/{_file}",
                        "save_path": WRFRUNConfig.WRF_WORK_PATH,
                        "save_name": _file,
                        "is_data": True,
                        "is_output": False
                    }
                    self.add_input_files(_file_config)

        super().before_exec()

        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WRF_WORK_PATH}/{NamelistName.WRF}", "wrf")

    def after_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:
            self.add_output_files(save_path=self._log_save_path, startswith="rsl.", outputs="namelist.input")
            self.add_output_files(save_path=self._output_save_path, startswith="wrfout")
            if self.save_restarts:
                restart_save_path = f"{self._output_save_path}/restart"
                self.add_output_files(save_path=restart_save_path, startswith="wrfrst", no_file_error=False)

        super().after_exec()

        logger.info(f"All wrf output files have been copied to {WRFRUNConfig.parse_resource_uri(self._output_save_path)}")


class DFI(ExecutableBase):
    """
    Execute "wrf.exe" to run DFI.
    """

    def __init__(self, input_file_dir_path: Optional[str] = None, update_real_output=True, core_num: Optional[int] = None):
        """
        Execute "wrf.exe" to run DFI.

        :param input_file_dir_path: Path of the directory that stores input data for "wrf.exe".
        :type input_file_dir_path: str
        :param update_real_output: If update the corresponding file in real.exe output directory.
        :type update_real_output: bool
        :param core_num: An positive integer number of used core numbers. ``mpirun`` will be used to execute geogrid.exe if ``core_num != None``.
        :type core_num: int
        """
        if isinstance(core_num, int) and core_num <= 0:
            logger.warning(f"`core_num` should be greater than 0")
            core_num = None

        if core_num is None:
            mpi_use = False
            mpi_cmd = None
            mpi_core_num = core_num

        else:
            mpi_use = True
            mpi_cmd = "mpirun"
            mpi_core_num = core_num

        super().__init__(name="dfi", cmd="./wrf.exe", work_path=WRFRUNConfig.WRF_WORK_PATH, mpi_use=mpi_use, mpi_cmd=mpi_cmd, mpi_core_num=mpi_core_num)

        self.input_file_dir_path = input_file_dir_path
        self.update_real_output = update_real_output

    def generate_custom_config(self):
        self.custom_config.update(
            {
                "input_file_dir_path": self.input_file_dir_path,
                "update_real_output": self.update_real_output,
                "namelist": WRFRUNConfig.get_namelist("dfi")
            }
        )

    def load_custom_config(self):
        self.input_file_dir_path = self.custom_config["input_file_dir_path"]
        self.update_real_output = self.custom_config["update_real_output"]

        if not WRFRUNConfig.register_custom_namelist_id("dfi"):
            logger.error("Can't register namelist for DFI.")
            raise NamelistError("Can't register namelist for DFI.")
        WRFRUNConfig.update_namelist(self.custom_config["namelist"], "dfi")

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        WRFRUNConfig.WRFRUN_WORK_STATUS = "dfi"

        if not WRFRUNConfig.IS_IN_REPLAY and not WRFRUNConfig.FAKE_SIMULATION_MODE:
            # prepare config
            if self.input_file_dir_path is None:
                self.input_file_dir_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/real"
                is_output = True

            else:
                is_output = False

            input_file_dir_path = WRFRUNConfig.parse_resource_uri(self.input_file_dir_path)

            if exists(input_file_dir_path):
                file_list = [x for x in listdir(input_file_dir_path) if x != "logs"]

                for _file in file_list:
                    _file_config: FileConfigDict = {
                        "file_path": f"{self.input_file_dir_path}/{_file}",
                        "save_path": WRFRUNConfig.WRF_WORK_PATH,
                        "save_name": _file,
                        "is_data": True,
                        "is_output": is_output
                    }
                    self.add_input_files(_file_config)

            if not WRFRUNConfig.register_custom_namelist_id("dfi"):
                logger.error("Can't register namelist for DFI.")
                raise NamelistError("Can't register namelist for DFI.")

            prepare_dfi_namelist()

        super().before_exec()
        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WRF_WORK_PATH}/{NamelistName.WRF}", "dfi")

    def after_exec(self):
        if not WRFRUNConfig.IS_IN_REPLAY:
            self.add_output_files(save_path=self._log_save_path, startswith="rsl.", outputs="namelist.input")
            self.add_output_files(save_path=self._output_save_path, startswith="wrfinput_initialized_")

        super().after_exec()

        parsed_output_save_path = WRFRUNConfig.parse_resource_uri(self._output_save_path)
        if self.update_real_output and not WRFRUNConfig.FAKE_SIMULATION_MODE:
            real_dir_path = WRFRUNConfig.parse_resource_uri(self.input_file_dir_path)

            move(f"{real_dir_path}/wrfinput_d01", f"{real_dir_path}/wrfinput_d01_before_dfi")
            copyfile(f"{parsed_output_save_path}/wrfinput_initialized_d01", f"{real_dir_path}/wrfinput_d01")
            logger.info(f"Replace real.exe output 'wrfinput_d01' with outputs, old file has been renamed as 'wrfinput_d01_before_dfi'")

        logger.info(f"All DFI output files have been copied to {parsed_output_save_path}")


class NDown(ExecutableBase):
    """
    Execute "ndown.exe".
    """

    def __init__(self, wrfout_file_path: str, real_output_dir_path: Optional[str] = None, update_namelist=True, core_num: Optional[int] = None):
        """
        Execute "ndown.exe".

        :param wrfout_file_path: wrfout file path.
        :type wrfout_file_path: str
        :param real_output_dir_path: Path of the directory that contains output of "real.exe".
        :type real_output_dir_path: str
        :param update_namelist: If update wrf's namelist for the final integral.
        :type update_namelist: bool
        :param core_num:
        :type core_num:
        """
        if isinstance(core_num, int) and core_num <= 0:
            logger.warning(f"`core_num` should be greater than 0")
            core_num = None

        if core_num is None:
            mpi_use = False
            mpi_cmd = None
            mpi_core_num = core_num

        else:
            mpi_use = True
            mpi_cmd = "mpirun"
            mpi_core_num = core_num

        _check_namelist_preparation()

        super().__init__(name="ndown", cmd="./ndown.exe", work_path=WRFRUNConfig.WRF_WORK_PATH, mpi_use=mpi_use, mpi_cmd=mpi_cmd, mpi_core_num=mpi_core_num)

        self.wrfout_file_path = wrfout_file_path
        self.real_output_dir_path = real_output_dir_path
        self.update_namelist = update_namelist

    def generate_custom_config(self):
        self.class_config["class_args"] = (self.wrfout_file_path,)
        self.custom_config.update(
            {
                "real_output_dir_path": self.real_output_dir_path,
                "update_namelist": self.update_namelist,
                "namelist": WRFRUNConfig.get_namelist("wrf"),
            }
        )

    def load_custom_config(self):
        self.real_output_dir_path = self.custom_config["real_output_dir_path"]
        self.update_namelist = self.custom_config["update_namelist"]
        WRFRUNConfig.update_namelist(self.custom_config["namelist"])

    def before_exec(self):
        WRFRUNConfig.check_wrfrun_context(True)
        WRFRUNConfig.WRFRUN_WORK_STATUS = "ndown"

        # we need to make sure time_control.io_form_auxinput2 is 2.
        # which means the format of input stream 2 is NetCDF.
        WRFRUNConfig.update_namelist({"time_control": {"io_form_auxinput2": 2}}, "wrf")

        if self.real_output_dir_path is None:
            self.real_output_dir_path = f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/real"
            is_output = True

        else:
            is_output = False

        wrfndi_file_config: FileConfigDict = {
            "file_path": f"{self.real_output_dir_path}/wrfinput_d02",
            "save_path": WRFRUNConfig.WRF_WORK_PATH,
            "save_name": "wrfndi_d02",
            "is_data": True,
            "is_output": is_output
        }
        wrfout_file_config: FileConfigDict = {
            "file_path": self.wrfout_file_path,
            "save_path": WRFRUNConfig.WRF_WORK_PATH,
            "save_name": "wrfout_d01",
            "is_data": True,
            "is_output": False
        }

        self.add_input_files([wrfndi_file_config, wrfout_file_config])

        super().before_exec()

        WRFRUNConfig.write_namelist(f"{WRFRUNConfig.WRF_WORK_PATH}/{NamelistName.WRF}", "wrf")

    def after_exec(self):
        self.add_output_files(save_path=self._log_save_path, startswith="rsl.", outputs="namelist.input")
        self.add_output_files(save_path=self._output_save_path, outputs=["wrfinput_d02", "wrfbdy_d02"])
        # also save other outputs of real.exe, so WRF can directly use them.
        self.add_output_files(output_dir=f"{WRFRUNConfig.WRFRUN_OUTPUT_PATH}/real", save_path=self._output_save_path,
                              startswith="wrflowinp_", no_file_error=False)

        super().after_exec()

        parsed_output_save_path = WRFRUNConfig.parse_resource_uri(self._output_save_path)

        move(f"{parsed_output_save_path}/wrfinput_d02", f"{parsed_output_save_path}/wrfinput_d01")
        move(f"{parsed_output_save_path}/wrfbdy_d02", f"{parsed_output_save_path}/wrfbdy_d01")

        logger.info(f"All ndown output files have been copied to {parsed_output_save_path}")

        if self.update_namelist:
            process_after_ndown()


class_list = [GeoGrid, LinkGrib, UnGrib, MetGrid, Real, WRF, NDown]
class_id_list = ["geogrid", "link_grib", "ungrib", "metgrid", "real", "wrf", "ndown"]

for _class, _id in zip(class_list, class_id_list):
    if not WRFRUNExecDB.is_registered(_id):
        WRFRUNExecDB.register_exec(_id, _class)

__all__ = ["GeoGrid", "LinkGrib", "UnGrib", "MetGrid", "Real", "WRF", "DFI", "NDown"]



from json import loads, dumps
from typing import Union, Tuple
from zipfile import ZipFile

from pandas import DataFrame, read_csv
import numpy as np

from wrfrun.core import WRFRUNConfig
from wrfrun.utils import logger


def to_fstring(var: Union[int, float, bool, str], length: Union[int, Tuple[int, int]]) -> str:
    """
    Convert a basic variable to a string following the Fortran standard.

    :param var: Basic variable that can be one of the ``[int, float, bool, str]``.
    :type var: Union[int, float, bool, str]
    :param length: The length of the output string. If the type of ``var`` is ``float``, the length must contain two parameters ``(total length, decimal length)``.
    :type length: Union[int, Tuple[int, int]]
    :return: Converted string.
    :rtype: str
    """
    if isinstance(var, float):
        if not isinstance(length, tuple):
            logger.error(
                "`length` must be a tuple contain two values `(total length, decimal length)` when `var` is `float`"
            )
            raise ValueError(
                "`length` must be a tuple contain two values `(total length, decimal length)` when `var` is `float`"
            )

        res = f"{var:{length[0]}.{length[1]}}"

    else:
        if not isinstance(length, int):
            logger.error(
                "`length` must be an int value when `var` is not `float`"
            )
            raise ValueError(
                "`length` must be an int value when `var` is not `float`"
            )

        if isinstance(var, bool):
            res = "T" if var else "F"
            res = res.rjust(length, " ")

        else:
            res = f"{var:{length}}"

    return res


class LittleRHead(dict):
    """
    Head info class for LittleR format data.
    """
    def __init__(
        self,
        longitude: float,
        latitude: float,
        fm: str,
        elevation: float,
        is_bogus: bool,
        date: str,
        ID="-----",
        name="wrfrun",
        source="Created by wrfrun package",
        num_sequence=0,
        sea_level_pressure=-888888,
        reference_pressure=-888888,
        surface_pressure=-888888,
        cloud_cover=-888888,
        precipitable_water=-888888,
        quality_control: Union[dict, int] = 0,
        **kwargs
    ) -> None:
        """Head info of little_r obs data.

        :param longitude: Longitude.
        :type longitude: float
        :param latitude: Latitude.
        :type latitude: float
        :param fm: Platform code (FM-Code).
        :type fm: str
        :param elevation: Elevation.
        :type elevation: float
        :param is_bogus: Is bogus?
        :type is_bogus: bool
        :param date: Date string in format ``%Y%m%d%H%M%S``, UTC time.
        :type date: str
        :param ID: ID, defaults to "-----".
        :type ID: str
        :param name: Name, defaults to "wrfrun".
        :type name: str
        :param source: Data source, defaults to "Created by wrfrun package".
        :type source: str
        :param num_sequence: Sequence number which is used to merge multiple record data, the less the num is, the newer the data is. Defaults to 0.
        :type num_sequence: int
        :param sea_level_pressure: Sea level pressure. Defaults to -888,888.
        :type sea_level_pressure: int
        :param reference_pressure: Reference pressure. Defaults to -888,888.
        :type reference_pressure: int
        :param surface_pressure: Surface pressure. Defaults to -888,888.
        :type surface_pressure: int
        :param cloud_cover: Cloud cover. Defaults to -888,888.
        :type cloud_cover: int
        :param precipitable_water: Precipitable water. Defaults to -888,888.
        :type precipitable_water: int
        :param quality_control: Quality control flag for observation data that can be specified per data in a dict. Defaults to 0.
        :type quality_control: Union[dict, int]
        """
        super().__init__(
            longitude=longitude,
            latitude=latitude,
            fm=fm,
            elevation=elevation,
            is_bogus=is_bogus,
            date=date,
            ID=ID,
            name=name,
            source=source,
            num_sequence=num_sequence,
            sea_level_pressure=sea_level_pressure,
            reference_pressure=reference_pressure,
            surface_pressure=surface_pressure,
            cloud_cover=cloud_cover,
            precipitable_water=precipitable_water,
            quality_control=quality_control,
            num_valid_field=256,
            num_error=-888888,
            num_warning=-888888,
            num_duplicate=-888888,
            is_sounding=True,
            discard=False,
            time=-888888,
            julian_day=-888888,
            ground_temperature=-888888,
            sst=-888888,
            precipitation=-888888,
            temp_daily_max=-888888,
            temp_daily_min=-888888,
            temp_night_min=-888888,
            pressure_delta_3h=-888888,
            pressure_delta_24h=-888888,
            ceiling=-888888,
        )
        self.longitude = longitude
        self.latitude = latitude
        self.fm = fm
        self.elevation = elevation
        self.is_bogus = is_bogus
        self.date = date
        self.ID = ID
        self.name = name
        self.source = source
        self.num_sequence = num_sequence
        self.sea_level_pressure = sea_level_pressure
        self.reference_pressure = reference_pressure
        self.surface_pressure = surface_pressure
        self.cloud_cover = cloud_cover
        self.precipitable_water = precipitable_water
        self.quality_control = quality_control

        # other unused field
        self.num_valid_field = 256
        self.num_error = -888888
        self.num_warning = -888888
        self.num_duplicate = -888888
        self.is_sounding = True
        self.discard = False
        self.time = -888888
        self.julian_day = -888888
        self.ground_temperature = -888888
        self.sst = -888888
        self.precipitation = -888888
        self.temp_daily_max = -888888
        self.temp_daily_min = -888888
        self.temp_night_min = -888888
        self.pressure_delta_3h = -888888
        self.pressure_delta_24h = -888888
        self.ceiling = -888888

        _ = kwargs

    def __str__(self) -> str:
        return self._convert_to_fstring()

    # def __repr__(self) -> str:
    #     return self._convert_to_fstring()

    def _convert_to_fstring(self) -> str:
        return f"{self.latitude:20.5f}" \
               f"{self.longitude:20.5f}" \
               f"{self.ID.rjust(40, ' ')}" \
               f"{self.name.rjust(40, ' ')}" \
               f"{self.fm.rjust(40, ' ')}" \
               f"{self.source.rjust(40, ' ')}" \
               f"{self.elevation:20.5f}" \
               f"{self.num_valid_field:10d}" \
               f"{self.num_error:10d}" \
               f"{self.num_warning:10d}" \
               f"{self.num_sequence:10d}" \
               f"{self.num_duplicate:10d}" \
               f"{to_fstring(self.is_sounding, 10)}" \
               f"{to_fstring(self.is_bogus, 10)}" \
               f"{to_fstring(self.discard, 10)}" \
               f"{self.time:10d}" \
               f"{self.julian_day:10d}" \
               f"{self.date.rjust(20, ' ')}" + self._generate_data_qc()

    def _generate_data_qc(self) -> str:
        field = [
            "sea_level_pressure",
            "reference_pressure",
            "ground_temperature",
            "sst",
            "surface_pressure",
            "precipitation",
            "temp_daily_max",
            "temp_daily_min",
            "temp_night_min",
            "pressure_delta_3h",
            "pressure_delta_24h",
            "cloud_cover",
            "ceiling",
            "precipitable_water",
        ]

        if isinstance(self.quality_control, int):
            res = ""
            for key in field:
                res += f"{getattr(self, key):13.5f}{self.quality_control:7d}"
        else:
            res = ""
            for key in field:
                res += f"{getattr(self, key):13.5f}{self.quality_control[key]:7d}"

        return res


LITTLE_R_DATA_FIELD = [
    "pressure", "pressure_qc",
    "height", "height_qc",
    "temperature", "temperature_qc",
    "dew_point", "dew_point_qc",
    "wind_speed", "wind_speed_qc",
    "wind_direction", "wind_direction_qc",
    "wind_u", "wind_u_qc",
    "wind_v", "wind_v_qc",
    "relative_humidity", "relative_humidity_qc",
    "thickness", "thickness_qc",
]


class LittleRData(DataFrame):
    def __init__(
        self,
        data=None,
        index=None,
        columns=None,
        pressure: Union[np.ndarray, float] = 100000.,
        height: Union[np.ndarray, float] = -888888.,
        temperature: Union[np.ndarray, float] = 264.,
        dew_point: Union[np.ndarray, float] = 263.,
        wind_speed: Union[np.ndarray, float] = -888888.,
        wind_direction: Union[np.ndarray, float] = -888888.,
        wind_u: Union[np.ndarray, float] = -888888.,
        wind_v: Union[np.ndarray, float] = -888888.,
        relative_humidity: Union[np.ndarray, float] = -888888.,
        thickness: Union[np.ndarray, float] = -888888.,
        quality_control_flag: Union[np.ndarray, dict, int] = 0,
        **kwargs
    ) -> None:
        # check data type
        if data is not None:
            super().__init__(data=data, index=index, columns=columns, **kwargs)  # type: ignore

        else:
            if isinstance(pressure, float):
                pressure = np.array([pressure]).astype(float)
                height = np.array([height]).astype(float)
                temperature = np.array([temperature]).astype(float)
                dew_point = np.array([dew_point]).astype(float)
                wind_speed = np.array([wind_speed]).astype(float)
                wind_direction = np.array([wind_direction]).astype(float)
                wind_u = np.array([wind_u]).astype(float)
                wind_v = np.array([wind_v]).astype(float)
                relative_humidity = np.array([relative_humidity]).astype(float)
                thickness = np.array([thickness]).astype(float)
                quality_control_flag = np.array(
                    [quality_control_flag]
                ).astype(int)
            else:
                pressure = np.asarray(pressure).astype(float)
                height = np.asarray(height).astype(float)
                temperature = np.asarray(temperature).astype(float)
                dew_point = np.asarray(dew_point).astype(float)
                wind_speed = np.asarray(wind_speed).astype(float)
                wind_direction = np.asarray(wind_direction).astype(float)
                wind_u = np.asarray(wind_u).astype(float)
                wind_v = np.asarray(wind_v).astype(float)
                relative_humidity = np.asarray(relative_humidity).astype(float)
                thickness = np.asarray(thickness).astype(float)
                quality_control_flag = np.asarray(
                    quality_control_flag
                ).astype(int)

            # construct data
            if isinstance(quality_control_flag, dict):
                data = {
                    "pressure": pressure,
                    "pressure_qc": quality_control_flag["pressure"],
                    "height": height,
                    "height_qc": quality_control_flag["height"],
                    "temperature": temperature,
                    "temperature_qc": quality_control_flag["temperature"],
                    "dew_point": dew_point,
                    "dew_point_qc": quality_control_flag["dew_point"],
                    "wind_speed": wind_speed,
                    "wind_speed_qc": quality_control_flag["wind_speed"],
                    "wind_direction": wind_direction,
                    "wind_direction_qc": quality_control_flag["wind_direction"],
                    "wind_u": wind_u,
                    "wind_u_qc": quality_control_flag["wind_u"],
                    "wind_v": wind_v,
                    "wind_v_qc": quality_control_flag["wind_v"],
                    "relative_humidity": relative_humidity,
                    "relative_humidity_qc": quality_control_flag["relative_humidity"],
                    "thickness": thickness,
                    "thickness_qc": quality_control_flag["thickness"],
                }
            else:
                data = {
                    "pressure": pressure,
                    "pressure_qc": quality_control_flag,
                    "height": height,
                    "height_qc": quality_control_flag,
                    "temperature": temperature,
                    "temperature_qc": quality_control_flag,
                    "dew_point": dew_point,
                    "dew_point_qc": quality_control_flag,
                    "wind_speed": wind_speed,
                    "wind_speed_qc": quality_control_flag,
                    "wind_direction": wind_direction,
                    "wind_direction_qc": quality_control_flag,
                    "wind_u": wind_u,
                    "wind_u_qc": quality_control_flag,
                    "wind_v": wind_v,
                    "wind_v_qc": quality_control_flag,
                    "relative_humidity": relative_humidity,
                    "relative_humidity_qc": quality_control_flag,
                    "thickness": thickness,
                    "thickness_qc": quality_control_flag,
                }

            super().__init__(data=data)  # type: ignore

    @classmethod
    def from_csv(cls, csv_path):
        """
        Read saved LittleR data from a CSV file.

        :param csv_path: CSV file path.
        :type csv_path: str
        :return:
        :rtype:
        """
        data_dict = read_csv(csv_path).to_dict()
        return cls.from_dict(data_dict)

    @classmethod
    def from_dict(cls, data: dict, orient='columns', dtype=None, columns=None):
        # check fields
        data_key = next(iter(data))
        temp_data = data[data_key]
        for field in LITTLE_R_DATA_FIELD:
            if field not in data:
                if field.endswith("_qc"):
                    data[field] = np.zeros_like(temp_data).astype(int)
                else:
                    data[field] = np.zeros_like(temp_data) - 888888.

        return super().from_dict(data, orient, dtype, columns)  # type: ignore

    def __str__(self) -> str:
        return self._convert_to_fstring()

    def _convert_to_fstring(self) -> str:
        fields = [
            "pressure",
            "height",
            "temperature",
            "dew_point",
            "wind_speed",
            "wind_direction",
            "wind_u",
            "wind_v",
            "relative_humidity",
            "thickness",
        ]
        qc_fields = [
            "pressure_qc",
            "height_qc",
            "temperature_qc",
            "dew_point_qc",
            "wind_speed_qc",
            "wind_direction_qc",
            "wind_u_qc",
            "wind_v_qc",
            "relative_humidity_qc",
            "thickness_qc",
        ]

        res = ""
        valid_field_num = 0
        for row in self.index:
            for (key, qc_key) in zip(fields, qc_fields):
                _field = self.loc[row, key]
                _field_qc = self.loc[row, qc_key]

                if _field > -100:  # type: ignore
                    valid_field_num += 1

                res += f"{_field:13.5f}{_field_qc:7d}"
            res += "\n"

        # add ending record
        for (_, _) in zip(fields, qc_fields):
            res += f"{-777777:13.5f}{0:7d}"
        res += "\n"

        # add tail integers
        res += f"{valid_field_num:7d}{0:7d}{0:7d}"

        return res


class LittleR(LittleRData):

    _metadata = ["little_r_head"]

    def __init__(
        self,
        data=None,
        index=None,
        columns=None,
        data_header: Union[dict, None] = None,
        pressure: Union[np.ndarray, float] = 100000.,
        height: Union[np.ndarray, float] = -888888.,
        temperature: Union[np.ndarray, float] = 264.,
        dew_point: Union[np.ndarray, float] = 263.,
        wind_speed: Union[np.ndarray, float] = -888888.,
        wind_direction: Union[np.ndarray, float] = -888888.,
        wind_u: Union[np.ndarray, float] = -888888.,
        wind_v: Union[np.ndarray, float] = -888888.,
        relative_humidity: Union[np.ndarray, float] = -888888.,
        thickness: Union[np.ndarray, float] = -888888.,
        quality_control_flag: Union[np.ndarray, dict, int] = 0,
        **kwargs
    ) -> None:
        super().__init__(
            data=data,
            pressure=pressure,
            height=height,
            temperature=temperature,
            dew_point=dew_point,
            wind_speed=wind_speed,
            wind_direction=wind_direction,
            wind_u=wind_u,
            wind_v=wind_v,
            relative_humidity=relative_humidity,
            thickness=thickness,
            quality_control_flag=quality_control_flag,
            index=index,
            columns=columns,
            **kwargs
        )

        if data_header is None:
            self.little_r_head = None
        else:
            self.little_r_head = LittleRHead(**data_header)

    def set_header(
        self,
        longitude: float,
        latitude: float,
        fm: str,
        elevation: float,
        is_bogus: bool,
        date: str,
        ID="-----",
        name="wrfrun",
        source="Created by wrfrun package",
        num_sequence=0,
        sea_level_pressure=-888888,
        reference_pressure=-888888,
        surface_pressure=-888888,
        cloud_cover=-888888,
        precipitable_water=-888888,
        quality_control: Union[dict, int] = 0,
        **kwargs
    ):
        """Head info of little_r obs data.

        :param longitude: Longitude.
        :type longitude: float
        :param latitude: Latitude.
        :type latitude: float
        :param fm: Platform code (FM-Code).
        :type fm: str
        :param elevation: Elevation.
        :type elevation: float
        :param is_bogus: Is bogus?
        :type is_bogus: bool
        :param date: Date string in format ``%Y%m%d%H%M%S``, UTC time.
        :type date: str
        :param ID: ID, defaults to "-----".
        :type ID: str, optional
        :param name: Name, defaults to "wrfrun".
        :type name: str, optional
        :param source: Data source, defaults to "Created by wrfrun package".
        :type source: str, optional
        :param num_sequence: Sequence number, used to merge multiple record data, the less the num is, the newer the data is, defaults to 0
        :type num_sequence: int, optional
        :param sea_level_pressure: Sea level pressure, defaults to -888888
        :type sea_level_pressure: int, optional
        :param reference_pressure: Reference pressure, defaults to -888888
        :type reference_pressure: int, optional
        :param surface_pressure: Surface pressure, defaults to -888888
        :type surface_pressure: int, optional
        :param cloud_cover: Cloud cover, defaults to -888888
        :type cloud_cover: int, optional
        :param precipitable_water: Precipitable water, defaults to -888888
        :type precipitable_water: int, optional
        :param quality_control: Quality control flag for observation data, can be specified per data in a dict, defaults to 0
        :type quality_control: Union[dict, int], optional
        """
        self.little_r_head = LittleRHead(
            longitude, latitude, fm, elevation, is_bogus, date, ID,
            name=name, source=source, num_sequence=num_sequence, sea_level_pressure=sea_level_pressure, reference_pressure=reference_pressure,
            surface_pressure=surface_pressure, cloud_cover=cloud_cover, precipitable_water=precipitable_water, quality_control=quality_control,
            **kwargs
        )

    def __str__(self) -> str:
        data_str = super().__str__()
        return f"{str(self.little_r_head)}\n{data_str}"

    @classmethod
    def from_csv(cls, csv_path):
        return super().from_csv(csv_path)

    def to_zlr(self, file_path: str):
        """
        Save all LittleR data to a ".zlr" file.

        :param file_path: The save path.
        :type file_path: str
        :return:
        :rtype:
        """
        if not file_path.endswith(".zlr"):
            file_path = f"{file_path}.zlr"

        file_path = WRFRUNConfig.parse_resource_uri(file_path)

        with ZipFile(file_path, "w") as zip_file:
            with zip_file.open("header", "w") as header_file:
                header_file.write(dumps(self.little_r_head).encode())

            with zip_file.open("data", "w") as data_file:
                self.to_csv(data_file, index=False)

    @classmethod
    def from_zlr(cls, file_path: str):
        """
        Read data from a ".zlr" file.

        :param file_path: The file path.
        :type file_path: str
        :return:
        :rtype:
        """
        file_path = WRFRUNConfig.parse_resource_uri(file_path)

        with ZipFile(file_path, "r") as zip_file:
            with zip_file.open("header", "r") as header_file:
                header = loads(header_file.read().decode())

            with zip_file.open("data", "r") as data_file:
                little_r = cls.from_csv(data_file)  # type: ignore

        little_r.set_header(**header)  # type: ignore

        return little_r


__all__ = ["LittleRHead", "LittleRData", "LittleR"]



from dataclasses import dataclass


@dataclass
class NamelistName:
    """
    Namelist file names.
    """
    WPS = "namelist.wps"
    WRF = "namelist.input"
    WRFDA = "namelist.input"


__all__ = ["NamelistName"]



class WRFRunBasicError(Exception):
    """
    Basic error type.
    """
    pass


class ConfigError(WRFRunBasicError):
    """
    Config cannot be used error.
    """
    pass


class WRFRunContextError(WRFRunBasicError):
    """
    Need to enter wrfrun context.
    """
    pass


class CommandError(WRFRunBasicError):
    """
    The command is wrong.
    """
    pass


class LoadConfigError(WRFRunBasicError):
    """
    Failed to load config.
    """
    pass


class OutputFileError(WRFRunBasicError):
    """
    No output found.
    """
    pass


class ResourceURIError(WRFRunBasicError):
    """
    Error about resource namespace.
    """
    pass


class InputFileError(WRFRunBasicError):
    """
    Input file error.
    """
    pass


class NamelistError(WRFRunBasicError):
    """
    Error about namelist.
    """
    pass


class ExecRegisterError(WRFRunBasicError):
    pass


class GetExecClassError(WRFRunBasicError):
    pass


class ModelNameError(WRFRunBasicError):
    """
    Name of the model isn't found in the config file.
    """
    pass


__all__ = ["WRFRunBasicError", "ConfigError", "WRFRunContextError", "CommandError", "LoadConfigError", "OutputFileError", "ResourceURIError", "InputFileError",
           "NamelistError", "ExecRegisterError", "GetExecClassError", "ModelNameError"]



from os import listdir
from os.path import exists
from shutil import copyfile
from typing import List, Optional

from wrfrun.core import WRFRUNConfig
from wrfrun.utils import check_path, logger


def extension_postprocess(output_dir: str, extension_id: str, outputs: Optional[List[str]] = None):
    """
    Apply postprocess after running an extension.
    This function can help you save all files and logs in ``output_dir`` to the ``output_path/extension_id`` and ``output_path/extension_id/logs``,
    ``output_path`` is defined in ``config.yaml``.
    Files end with ``.log`` are treated as log files, while others are treated as outputs.
    You can specify outputs by giving their names through parameter ``outputs``.

    Save all outputs of ungrib.

    >>> from wrfrun.extension.utils import extension_postprocess
    >>> output_dir_path = "/WPS/outputs"
    >>> extension_postprocess(output_dir_path, "ungrib")

    Save outputs start with ``SST_FILE`` of ungrib.

    >>> from os import listdir
    >>> from wrfrun.extension.utils import extension_postprocess
    >>> output_dir_path = "/WPS/outputs"
    >>> outputs_name = [x for x in listdir(output_dir_path) if x.startswith("SST_FILE")]    # type: ignore
    >>> extension_postprocess(output_dir_path, "ungrib", outputs=outputs_name)

    :param output_dir: Absolute path of output directory.
    :param extension_id: A unique id to distinguish different extensions.
                         And all files and logs will be saved to ``output_path/extension_id`` and ``output_path/extension_id/logs``,
                         ``output_path`` is defined in ``config.yaml``.
    :param outputs: A list contains multiple filenames. Files in this will be treated as outputs.
    :return:
    """
    output_path = WRFRUNConfig.WRFRUN_OUTPUT_PATH
    output_save_path = f"{output_path}/{extension_id}"
    log_save_path = f"{output_path}/{extension_id}/logs"

    output_save_path = WRFRUNConfig.parse_resource_uri(output_save_path)
    log_save_path = WRFRUNConfig.parse_resource_uri(log_save_path)
    output_dir = WRFRUNConfig.parse_resource_uri(output_dir)

    filenames = listdir(output_dir)
    logs = [x for x in filenames if x.endswith(".log")]
    if outputs is None:
        outputs = list(set(filenames) - set(logs))

    check_path(output_save_path, log_save_path)

    logger.info(f"Saving outputs and logs to {output_save_path}")

    for _file in outputs:
        if exists(f"{output_dir}/{_file}"):
            copyfile(f"{output_dir}/{_file}", f"{output_save_path}/{_file}")
        else:
            logger.warning(f"Output {_file} not found")

    for _log in logs:
        copyfile(f"{output_dir}/{_log}", f"{log_save_path}/{_log}")


__all__ = ["extension_postprocess"]



import argparse
from json import loads
from os import listdir
from os.path import dirname, exists
from typing import Union

HEADER_CODE = """from os.path import abspath, dirname

from wrfrun.core import WRFRUNConfig


RES_PATH = abspath(dirname(__file__))
WRFRUNConfig.register_resource_uri(WRFRUNConfig.WRFRUN_RESOURCE_PATH, RES_PATH)

"""

FOOT_CODE = """\nWRFRUNConfig.set_config_template_path(CONFIG_TOML_TEMPLATE)\n"""


def _generate_init():
    parser = argparse.ArgumentParser(description="Generate __init__.py")
    parser.add_argument("--output", "-o", help="Absolute or relative path to '__init__.py'.", required=True)
    args = vars(parser.parse_args())

    target_file_path = args["output"]
    target_dir_path = dirname(target_file_path)

    total_name_list = []
    total_path_list = []
    file_name_list, file_path_list, dir_name_list, dir_path_list = _generate_name_list_in_dir(target_dir_path, "")
    total_name_list += file_name_list
    total_path_list += file_path_list

    while len(dir_name_list) > 0:
        file_name_list, file_path_list, dir_name_list, dir_path_list = _generate_name_list_in_dir(dir_path_list, dir_name_list)
        total_name_list += file_name_list
        total_path_list += file_path_list

    total_path_list = [f":WRFRUN_RESOURCE_PATH:/{x[len(target_dir_path) + 1:]}" for x in total_path_list]

    with open(target_file_path, "w") as f:
        f.write(HEADER_CODE)

        all_string = ""
        for _var_name, _file_path in zip(total_name_list, total_path_list):
            f.write(f"{_var_name} = \"{_file_path}\"\n")
            all_string += f', "{_var_name}"'

        f.write(FOOT_CODE)

        all_string = all_string.strip(", ")

        f.write(f'\n\n__all__ = [{all_string}]\n')


def _generate_name_list_in_dir(dir_path: Union[str, list[str]], dir_name: Union[str, list[str]], root=False) -> tuple[list[str], list[str], list[str], list[str]]:
    """
    Generate a variable name list for the target path.

    :param dir_path:
    :type dir_path:
    :param dir_name:
    :type dir_name:
    :param root: Root path?
    :type root: bool
    :return: (file_name_list, file_path_list, dir_name_list, dir_path_list).
    :rtype: tuple
    """
    if isinstance(dir_path, str):
        dir_path = [dir_path, ]
        dir_name = [dir_name, ]

    file_name_list = []
    file_path_list = []
    dir_name_list = []
    dir_path_list = []

    for _dir_path, _dir_name in zip(dir_path, dir_name):
        if not exists(f"{_dir_path}/name_map.json"):
            raise FileNotFoundError(f"{_dir_path}/name_map.json. Files: {listdir(_dir_path)}")

        with open(f"{_dir_path}/name_map.json", "r") as f:
            name_map_dict: dict[str, dict[str, str]] = loads(f.read())

        for file, config in name_map_dict.items():
            name = config["name"]
            file_type = config["type"]

            var_name = f"{_dir_name}_{name}"
            var_name = var_name.strip("_")

            if root:
                _path = f":WRFRUN_RESOURCE_PATH:/{file}"
            else:
                _path = f"{_dir_path}/{file}"

            if file_type == "dir":
                dir_name_list.append(var_name)
                dir_path_list.append(_path)

            else:
                file_name_list.append(var_name)
                file_path_list.append(_path)

    return file_name_list, file_path_list, dir_name_list, dir_path_list


_generate_init()
